# from __future__ import annotations  # breaks msgspec Annotated

from typing import ClassVar
from unittest.mock import Mock

import pytest

from psygnal import (
    Signal,
    SignalGroup,
    SignalGroupDescriptor,
)


@pytest.mark.parametrize(
    "type_",
    [
        "dataclass",
        "attrs",
        "pydantic",
        "msgspec",
    ],
)
def test_alias_parameters(type_: str) -> None:
    root_aliases = {"b": None, "bb": None}

    class MyGroup(SignalGroup, signal_aliases=root_aliases):
        b = Signal(str, str)
        bb = Signal(str, str)

    foo_options = {"signal_aliases": {"_b": None}}
    bar_options = {
        "signal_aliases": lambda x: None if x.startswith("_") else f"{x}_changed"
    }
    baz_options = {"signal_aliases": {"a": "a_changed", "_b": "b_changed"}}
    baz2_options = {
        "signal_group_class": MyGroup,
        "signal_aliases": {"aa": "a", "bb": "b"},
    }

    if type_ == "dataclass":
        from dataclasses import dataclass, field

        @dataclass
        class Foo:
            events: ClassVar = SignalGroupDescriptor(**foo_options)
            a: int
            _b: str

        @dataclass
        class Bar:
            events: ClassVar = SignalGroupDescriptor(**bar_options)
            a: int
            _b: str

        @dataclass
        class Baz:
            events: ClassVar = SignalGroupDescriptor(**baz_options)
            a: int
            _b: str = field(default="b")

            @property
            def b(self) -> str:
                return self._b

            @b.setter
            def b(self, value: str):
                self._b = value

        @dataclass
        class Baz2:
            events: ClassVar = SignalGroupDescriptor(**baz2_options)
            a: int
            aa: int
            b: str
            bb: str

    elif type_ == "attrs":
        from attrs import define, field

        @define
        class Foo:
            events: ClassVar = SignalGroupDescriptor(**foo_options)
            a: int
            _b: str = field(alias="_b")

        @define
        class Bar:
            events: ClassVar = SignalGroupDescriptor(**bar_options)
            a: int
            _b: str = field(alias="_b")

        @define
        class Baz:
            events: ClassVar = SignalGroupDescriptor(**baz_options)
            a: int
            _b: str = field(alias="_b", default="b")

            @property
            def b(self) -> str:
                return self._b

            @b.setter
            def b(self, value: str):
                self._b = value

        @define
        class Baz2:
            events: ClassVar = SignalGroupDescriptor(**baz2_options)
            a: int
            aa: int
            b: str
            bb: str

    elif type_ == "pydantic":
        pytest.importorskip("pydantic", minversion="2")
        from pydantic import BaseModel

        class Foo(BaseModel):
            events: ClassVar = SignalGroupDescriptor(**foo_options)
            a: int
            _b: str  # not a field anyway

        class Bar(BaseModel):
            events: ClassVar = SignalGroupDescriptor(**bar_options)
            a: int
            _b: str  # not a field anyway

        class Baz(BaseModel):
            events: ClassVar = SignalGroupDescriptor(**baz_options)
            a: int
            _b: str = "b"  # not defining a field, signal will not be created

            @property
            def b(self) -> str:
                return self._b

            @b.setter
            def b(self, value: str):
                self._b = value

        class Baz2(BaseModel):
            events: ClassVar = SignalGroupDescriptor(**baz2_options)
            a: int
            aa: int
            b: str
            bb: str

    elif type_ == "msgspec":
        msgspec = pytest.importorskip("msgspec")

        class Foo(msgspec.Struct):  # type: ignore
            events: ClassVar = SignalGroupDescriptor(**foo_options)
            a: int
            _b: str

        class Bar(msgspec.Struct):  # type: ignore
            events: ClassVar = SignalGroupDescriptor(**bar_options)
            a: int
            _b: str

        class Baz(msgspec.Struct):  # type: ignore
            events: ClassVar = SignalGroupDescriptor(**baz_options)
            a: int
            _b: str = "b"

            @property
            def b(self) -> str:
                return self._b

            @b.setter
            def b(self, value: str):
                self._b = value

        class Baz2(msgspec.Struct):  # type: ignore
            events: ClassVar = SignalGroupDescriptor(**baz2_options)
            a: int
            aa: int
            b: str
            bb: str

    # Instantiate objects
    foo = Foo(a=1, _b="b")
    bar = Bar(a=1, _b="b")
    baz = Baz(a=1)
    baz2 = Baz2(a=1, aa=2, b="b", bb="bb")

    # Check signals
    assert set(foo.events) == {"a"}
    assert hasattr(foo.events, "_psygnal_aliases")
    assert foo.events._psygnal_aliases == foo_options["signal_aliases"]

    assert set(bar.events) == {"a_changed"}
    assert hasattr(bar.events, "_psygnal_aliases")
    if type_.startswith("pydantic"):
        assert bar.events._psygnal_aliases == {"a": "a_changed"}
    else:
        assert bar.events._psygnal_aliases == {"a": "a_changed", "_b": None}

    if type_.startswith("pydantic"):
        assert set(baz.events) == {"a_changed"}
    else:
        assert set(baz.events) == {"a_changed", "b_changed"}
    assert hasattr(baz.events, "_psygnal_aliases")
    assert baz.events._psygnal_aliases == baz_options["signal_aliases"]

    # with pytest.warns(UserWarning, match=r"Skip signal \'a\', was already created"):
    with pytest.warns(UserWarning) as record:
        assert set(baz2.events) == {"a", "b", "bb"}
    assert len(record) == 2
    assert record[0].message.args[0].startswith("Skip signal 'a', was already created")
    assert record[1].message.args[0].startswith("Skip signal 'b', was already defined")
    assert hasattr(baz.events, "_psygnal_aliases")
    assert baz2.events._psygnal_aliases == {
        **root_aliases,
        **baz2_options["signal_aliases"],
    }

    mock = Mock()
    foo.events.a.connect(mock)
    bar.events.a_changed.connect(mock)
    baz.events.a_changed.connect(mock)
    if not type_.startswith("pydantic"):
        baz.events.b_changed.connect(mock)
    baz2.events.a.connect(mock)
    baz2.events.b.connect(mock)
    baz2.events.bb.connect(mock)

    # Foo
    foo.a = 1
    mock.assert_not_called()
    foo.a = 2
    mock.assert_called_once_with(2, 1)
    mock.reset_mock()
    foo._b = "b"
    foo._b = "c"
    mock.assert_not_called()

    # Bar
    bar.a = 1
    mock.assert_not_called()
    bar.a = 2
    mock.assert_called_once_with(2, 1)
    mock.reset_mock()
    bar._b = "b"
    bar._b = "c"
    mock.assert_not_called()

    # Baz
    baz.a = 1
    mock.assert_not_called()
    baz.a = 2
    mock.assert_called_once_with(2, 1)
    mock.reset_mock()

    # Baz2
    baz2.a = 1
    baz2.aa = 2
    mock.assert_not_called()
    baz2.a = 2
    mock.assert_called_once_with(2, 1)
    mock.reset_mock()
    baz2.aa = 3
    mock.assert_called_once_with(3, 2)
    mock.reset_mock()
    baz2.b = "b"
    mock.assert_not_called()
    baz2.b = "c"
    mock.assert_not_called()
    baz2.bb = "bb"
    mock.assert_not_called()
    baz2.bb = "bbb"
    mock.assert_called_once_with("bbb", "bb")
    mock.reset_mock()

    # pydantic v1 does not support properties
    if type_ == "pydantic_v1":
        return

    baz.b = "b"
    mock.assert_not_called()
    baz.b = "c"
    if not type_.startswith("pydantic"):
        mock.assert_called_once_with("c", "b")


def test_direct_signal_group() -> None:
    class FooSignalGroup(SignalGroup, signal_aliases={"e": None}):
        a = Signal(int, int)
        b_changed = Signal(float, float)
        c = Signal(str, str)
        d = Signal(str, str)
        e = Signal(str, str)

    class Foo:
        events: ClassVar = SignalGroupDescriptor(
            signal_group_class=FooSignalGroup,
            collect_fields=False,
            signal_aliases={
                "b": "b_changed",
                "c": None,
                "_c": "c",
                "_e": "e",
            },
        )
        a: int
        b: float
        _c: str
        _d: str
        _e: int

        def __init__(
            self,
            a: int = 1,
            b: float = 2.0,
            c: str = "c",
            d: str = "d",
            _e: int = 5,
        ):
            self.a = a
            self.b = b
            self.c = c
            self.d = d
            self._e = _e

        @property
        def c(self) -> str:
            return self._c

        @c.setter
        def c(self, value: str):
            self._c = value

        @property
        def d(self) -> str:
            return self._d.lower()

        @d.setter
        def d(self, value: str):
            self._d = value

    foo = Foo()

    assert hasattr(foo.events, "_psygnal_aliases")
    assert foo.events._psygnal_aliases == {
        "b": "b_changed",
        "_c": "c",
        "c": None,
        "e": None,
        "_e": "e",
    }

    mock = Mock()
    foo.events.a.connect(mock)
    foo.events.b_changed.connect(mock)
    foo.events.c.connect(mock)
    foo.events.d.connect(mock)
    foo.events.e.connect(mock)

    foo.events.e.emit("f", "e")
    mock.assert_called_once_with("f", "e")
    mock.reset_mock()

    foo.a = 2
    mock.assert_called_once_with(2, 1)
    mock.reset_mock()

    foo.b = 3.0
    mock.assert_called_once_with(3.0, 2.0)
    mock.reset_mock()

    foo.c = "c"
    mock.assert_not_called()
    foo.c = "cc"
    mock.assert_called_once_with("cc", "c")
    mock.reset_mock()
    foo._c = "ccc"
    mock.assert_called_once_with("ccc", "cc")
    mock.reset_mock()

    foo.d = "D"
    mock.assert_not_called()
    foo.d = "DD"
    mock.assert_called_once_with("dd", "d")
    mock.reset_mock()

    foo._e = 5
    mock.assert_not_called()
    foo._e = 6
    mock.assert_called_once_with(6, 5)
    mock.reset_mock()


def test_bad_siggroup_descriptor_init():
    with pytest.raises(
        TypeError,
        match="'signal_group_class' must be a subclass of SignalGroup",
    ):
        SignalGroupDescriptor(signal_group_class=type)  # type: ignore

    with pytest.raises(
        ValueError,
        match=r"Cannot use SignalGroup with `collect_fields=False`.",
    ):
        SignalGroupDescriptor(collect_fields=False)

    with pytest.raises(
        ValueError,
        match="Cannot use a Callable for `signal_aliases` with `collect_fields=False`",
    ):
        SignalGroupDescriptor(
            collect_fields=False,
            signal_group_class=type("MyGroup", (SignalGroup,), {"x": Signal()}),
            signal_aliases=lambda x: None,
        )



from unittest.mock import Mock, call

import numpy as np

from psygnal import EmissionInfo, SignalGroup
from psygnal.containers import (
    EventedCallableObjectProxy,
    EventedObjectProxy,
    _evented_proxy,
)
from psygnal.utils import monitor_events


def test_evented_proxy():
    class T:
        def __init__(self) -> None:
            self.x = 1
            self.f = "f"
            self._list = [0, 1]

        def __getitem__(self, key):
            return self._list[key]

        def __setitem__(self, key, value):
            self._list[key] = value

        def __delitem__(self, key):
            del self._list[key]

    t = EventedObjectProxy(T())
    assert "events" in dir(t)
    assert t.x == 1

    mock = Mock()
    with monitor_events(t.events, mock):
        t.x = 2
        t.f = "f"
        del t.x
        t.y = "new"
        t[0] = 7
        t[0] = 7  # no event
        del t[0]

    assert mock.call_args_list == [
        call(EmissionInfo(t.events.attribute_set, ("x", 2))),
        call(EmissionInfo(t.events.attribute_deleted, ("x",))),
        call(EmissionInfo(t.events.attribute_set, ("y", "new"))),
        call(EmissionInfo(t.events.item_set, (0, 7))),
        call(EmissionInfo(t.events.item_deleted, (0,))),
    ]


def test_evented_proxy_ref():
    class T:
        def __init__(self) -> None:
            self.x = 1

    assert not _evented_proxy._OBJ_CACHE
    t = EventedObjectProxy(T())
    assert not _evented_proxy._OBJ_CACHE
    assert isinstance(t.events, SignalGroup)  # this will actually create the group
    assert len(_evented_proxy._OBJ_CACHE) == 1

    del t  # this should clean up the object from the cache
    assert not _evented_proxy._OBJ_CACHE


def test_in_place_proxies():
    # fmt: off
    class T:
        x = 0
        def __iadd__(self, other): return self
        def __isub__(self, other): return self
        def __imul__(self, other): return self
        def __imatmul__(self, other): return self
        def __itruediv__(self, other): return self
        def __ifloordiv__(self, other): return self
        def __imod__(self, other): return self
        def __ipow__(self, other): return self
        def __ilshift__(self, other): return self
        def __irshift__(self, other): return self
        def __iand__(self, other): return self
        def __ixor__(self, other): return self
        def __ior__(self, other): return self
    # fmt: on

    t = EventedObjectProxy(T())
    mock = Mock()
    with monitor_events(t.events, mock):
        t += 1
        mock.assert_called_with(EmissionInfo(t.events.in_place, ("add", 1)))
        t -= 2
        mock.assert_called_with(EmissionInfo(t.events.in_place, ("sub", 2)))
        t *= 3
        mock.assert_called_with(EmissionInfo(t.events.in_place, ("mul", 3)))
        t /= 4
        mock.assert_called_with(EmissionInfo(t.events.in_place, ("truediv", 4)))
        t //= 5
        mock.assert_called_with(EmissionInfo(t.events.in_place, ("floordiv", 5)))
        t @= 6
        mock.assert_called_with(EmissionInfo(t.events.in_place, ("matmul", 6)))
        t %= 7
        mock.assert_called_with(EmissionInfo(t.events.in_place, ("mod", 7)))
        t **= 8
        mock.assert_called_with(EmissionInfo(t.events.in_place, ("pow", 8)))
        t <<= 9
        mock.assert_called_with(EmissionInfo(t.events.in_place, ("lshift", 9)))
        t >>= 10
        mock.assert_called_with(EmissionInfo(t.events.in_place, ("rshift", 10)))
        t &= 11
        mock.assert_called_with(EmissionInfo(t.events.in_place, ("and", 11)))
        t ^= 12
        mock.assert_called_with(EmissionInfo(t.events.in_place, ("xor", 12)))
        t |= 13
        mock.assert_called_with(EmissionInfo(t.events.in_place, ("or", 13)))


def test_numpy_proxy() -> None:
    ary = np.ones((4, 4))
    t: EventedObjectProxy[np.ndarray] = EventedObjectProxy(ary)
    assert repr(t) == repr(ary)

    mock = Mock()
    with monitor_events(t.events, mock):
        t[0] = 2
        info = next(iter(mock.call_args))[0]
        assert isinstance(info, EmissionInfo)
        signal = info.signal
        (key, value) = info.args

        assert signal.name == "item_set"
        assert key == 0
        assert np.array_equal(value, [2, 2, 2, 2])
        mock.reset_mock()

        t[2:] = np.arange(8).reshape(2, 4)

        info = next(iter(mock.call_args))[0]
        assert isinstance(info, EmissionInfo)
        signal = info.signal
        (key, value) = info.args

        assert signal.name == "item_set"
        assert key == slice(2, None, None)
        assert np.array_equal(value, [[0, 1, 2, 3], [4, 5, 6, 7]])
        mock.reset_mock()

        t += 1
        t -= 1
        assert np.array_equal(
            t, np.asarray([[2, 2, 2, 2], [1, 1, 1, 1], [0, 1, 2, 3], [4, 5, 6, 7]])
        )
        t *= [0, 0, 0, 0]
        assert not t.any()

    assert mock.call_args_list == [
        call(EmissionInfo(t.events.in_place, ("add", 1))),
        call(EmissionInfo(t.events.in_place, ("sub", 1))),
        call(EmissionInfo(t.events.in_place, ("mul", [0, 0, 0, 0]))),
    ]


def test_evented_callable_proxy():
    calls = []

    def f(*args, **kwargs):
        calls.append((args, kwargs))

    ef = EventedCallableObjectProxy(f)
    ef(1, 2, foo="bar")
    assert calls == [((1, 2), {"foo": "bar"})]



"""qtbot should work for testing!"""

from collections.abc import Callable
from threading import Thread, current_thread, main_thread
from typing import TYPE_CHECKING, Any, Literal
from unittest.mock import Mock

import pytest

from psygnal import Signal
from psygnal._signal import _guess_qtsignal_signature

pytest.importorskip("pytestqt")
if TYPE_CHECKING:
    from pytestqt.qtbot import QtBot


def _equals(*val: Any) -> Callable[[tuple[Any, ...]], bool]:
    def _inner(*other: Any) -> bool:
        return other == val

    return _inner


def test_wait_signals(qtbot: "QtBot") -> None:
    class Emitter:
        sig1 = Signal()
        sig2 = Signal(int)
        sig3 = Signal(int, int)

    e = Emitter()

    with qtbot.waitSignal(e.sig2, check_params_cb=_equals(1)):
        e.sig2.emit(1)

    with qtbot.waitSignal(e.sig3, check_params_cb=_equals(2, 3)):
        e.sig3.emit(2, 3)

    with qtbot.waitSignals([e.sig3], check_params_cbs=[_equals(2, 3)]):
        e.sig3.emit(2, 3)

    signals = [e.sig1, e.sig2, e.sig3, e.sig1]
    checks = [_equals(), _equals(1), _equals(2, 3), _equals()]
    with qtbot.waitSignals(signals, check_params_cbs=checks, order="strict"):
        e.sig1.emit()
        e.sig2.emit(1)
        e.sig3.emit(2, 3)
        e.sig1.emit()


def test_guess_signal_sig(qtbot: "QtBot") -> None:
    from qtpy import QtCore

    class QtObject(QtCore.QObject):
        qsig1 = QtCore.Signal()
        qsig2 = QtCore.Signal(int)
        qsig3 = QtCore.Signal(int, str)

    q_obj = QtObject()
    assert "qsig1()" in _guess_qtsignal_signature(q_obj.qsig1)
    assert "qsig1()" in _guess_qtsignal_signature(q_obj.qsig1.emit)
    assert "qsig2(int)" in _guess_qtsignal_signature(q_obj.qsig2)
    assert "qsig2(int)" in _guess_qtsignal_signature(q_obj.qsig2.emit)
    assert "qsig3(int,QString)" in _guess_qtsignal_signature(q_obj.qsig3)
    assert "qsig3(int,QString)" in _guess_qtsignal_signature(q_obj.qsig3.emit)


def test_connect_qt_signal_instance(qtbot: "QtBot") -> None:
    from qtpy import QtCore

    class Emitter:
        sig1 = Signal()
        sig2 = Signal(int)
        sig3 = Signal(int, int)

    class QtObject(QtCore.QObject):
        qsig1 = QtCore.Signal()
        qsig2 = QtCore.Signal(int)

    q_obj = QtObject()
    e = Emitter()

    # the hard case: signal.emit takes less args than we emit
    def test_receives_1(value: int) -> bool:
        # making sure that qsig2.emit only receives and emits 1 value
        return value == 1

    e.sig3.connect(q_obj.qsig2.emit)
    with qtbot.waitSignal(q_obj.qsig2, check_params_cb=test_receives_1):
        e.sig3.emit(1, 2)  # too many

    # the "standard" cases, where params match
    e.sig1.connect(q_obj.qsig1.emit)
    with qtbot.waitSignal(q_obj.qsig1):
        e.sig1.emit()

    e.sig2.connect(q_obj.qsig2.emit)
    with qtbot.waitSignal(q_obj.qsig2):
        e.sig2.emit(1)

    # the flip case: signal.emit takes more args than we emit
    with pytest.raises(ValueError):
        e.sig1.connect(q_obj.qsig2.emit)
    e.sig1.emit()


@pytest.mark.parametrize("thread", [None, "main"])
def test_q_main_thread_emit(
    thread: Literal["main", None], qtbot: "QtBot", qapp
) -> None:
    """Test using signal.emit(..., queue=True)

    ... and receiving it on the main thread with a QTimer connected to `emit_queued`
    """
    from psygnal.qt import start_emitting_from_queue, stop_emitting_from_queue

    class C:
        sig = Signal(int)

    obj = C()
    mock = Mock()

    @obj.sig.connect(thread=thread)
    def _some_slot(val: int) -> None:
        mock(val)
        assert (current_thread() == main_thread()) == (thread == "main")

    def _emit_from_thread() -> None:
        assert current_thread() != main_thread()
        obj.sig.emit(1)

    with qtbot.waitSignal(obj.sig, timeout=1000):
        t = Thread(target=_emit_from_thread)
        t.start()
        t.join()

    qapp.processEvents()
    if thread is None:
        mock.assert_called_once_with(1)
    else:
        mock.assert_not_called()
        start_emitting_from_queue()
        qapp.processEvents()
        mock.assert_called_once_with(1)

        start_emitting_from_queue(10)  # just for test coverage
        stop_emitting_from_queue()



import inspect
import sys
from collections.abc import Sequence
from contextlib import nullcontext
from typing import Any, ClassVar, Protocol, runtime_checkable
from unittest.mock import Mock, call, patch

import numpy as np
import pytest
from pydantic import Field

from psygnal import EventedModel
from psygnal.containers import EventedList
from psygnal.containers._evented_dict import EventedDict
from psygnal.containers._evented_set import EventedSet

try:
    from pydantic import PrivateAttr
except ImportError:
    pytest.skip("pydantic not installed", allow_module_level=True)

import pydantic.version
from pydantic import BaseModel

from psygnal import EmissionInfo
from psygnal._group import SignalGroup
from psygnal._signal import ReemissionMode

PYDANTIC_V2 = pydantic.version.VERSION.startswith("2")

try:
    from pydantic import field_serializer
except ImportError:

    def field_serializer(*args, **kwargs):
        def decorator(cls):
            return cls

        return decorator


def asdict(obj: "BaseModel") -> dict:
    if PYDANTIC_V2:
        return obj.model_dump()
    else:
        return obj.dict()


def asjson(obj: BaseModel) -> str:
    if PYDANTIC_V2:
        return obj.model_dump_json()
    else:
        return obj.json()


def test_creating_empty_evented_model():
    """Test creating an empty evented pydantic model."""
    model = EventedModel()
    assert model is not None
    assert model.events is not None


def test_evented_model():
    """Test creating an evented pydantic model."""

    class User(EventedModel):
        id: int
        name: str = "Alex"
        age: ClassVar[int] = 100

    user = User(id=0)
    # test basic functionality
    assert user.id == 0
    assert user.name == "Alex"

    user.id = 2
    assert user.id == 2

    # test event system
    assert isinstance(user.events, SignalGroup)
    assert "id" in user.events
    assert "name" in user.events

    # ClassVars are excluded from events
    assert "age" not in user.events

    id_mock = Mock()
    name_mock = Mock()
    user.events.id.connect(id_mock)
    user.events.name.connect(name_mock)
    # setting an attribute should, by default, emit an event with the value
    user.id = 4
    id_mock.assert_called_with(4)
    name_mock.assert_not_called()
    # and event should only be emitted when the value has changed.
    id_mock.reset_mock()
    user.id = 4
    id_mock.assert_not_called()
    name_mock.assert_not_called()


def test_evented_model_array_updates():
    """Test updating an evented pydantic model with an array."""

    class Model(EventedModel):
        """Demo evented model."""

        values: np.ndarray

        if PYDANTIC_V2:
            model_config = {"arbitrary_types_allowed": True}
        else:

            class Config:
                arbitrary_types_allowed = True

    first_values = np.array([1, 2, 3])
    model = Model(values=first_values)

    # Mock events
    values_mock = Mock()
    model.events.values.connect(values_mock)

    np.testing.assert_almost_equal(model.values, first_values)

    # Updating with new data
    new_array = np.array([1, 2, 4])
    model.values = new_array
    np.testing.assert_array_equal(values_mock.call_args.args[0], new_array)
    values_mock.reset_mock()

    # Updating with same data, no event should be emitted
    model.values = new_array
    values_mock.assert_not_called()


def test_evented_model_np_array_equality():
    """Test checking equality with an evented model with direct numpy."""

    class Model(EventedModel):
        values: np.ndarray

        if PYDANTIC_V2:
            model_config = {"arbitrary_types_allowed": True}
        else:

            class Config:
                arbitrary_types_allowed = True

    model1 = Model(values=np.array([1, 2, 3]))
    model2 = Model(values=np.array([1, 5, 6]))

    assert model1 == model1
    assert model1 != model2

    model2.values = np.array([1, 2, 3])
    assert model1 == model2


def test_evented_model_da_array_equality():
    """Test checking equality with an evented model with direct dask."""
    da = pytest.importorskip("dask.array")

    class Model(EventedModel):
        values: da.Array

        if PYDANTIC_V2:
            model_config = {"arbitrary_types_allowed": True}
        else:

            class Config:
                arbitrary_types_allowed = True

    r = da.ones((64, 64))
    model1 = Model(values=r)
    model2 = Model(values=da.ones((64, 64)))

    assert model1 == model1
    # dask arrays will only evaluate as equal if they are the same object.
    assert model1 != model2

    model2.values = r
    assert model1 == model2


def test_values_updated() -> None:
    class User(EventedModel):
        """Demo evented model.

        Parameters
        ----------
        id : int
            User id.
        name : str, optional
            User name.
        """

        id: int
        user_name: str = "A"
        age: ClassVar[int] = 100

    user1 = User(id=0)
    user2 = User(id=1, user_name="K")
    # Check user1 and user2 dicts
    assert asdict(user1) == {"id": 0, "user_name": "A"}
    assert asdict(user2) == {"id": 1, "user_name": "K"}

    # Add mocks
    user1_events = Mock()
    u1_id_events = Mock()
    u2_id_events = Mock()

    user1.events.all.connect(user1_events)
    user1.events.all.connect(user1_events)

    user1.events.id.connect(u1_id_events)
    user2.events.id.connect(u2_id_events)
    user1.events.id.connect(u1_id_events)
    user2.events.id.connect(u2_id_events)

    # Update user1 from user2
    user1.update(user2)
    assert asdict(user1) == {"id": 1, "user_name": "K"}

    u1_id_events.assert_called_with(1)
    u2_id_events.assert_not_called()

    # NOTE:
    # user.events.user_name is NOT actually emitted because it has no callbacks
    # connected to it.  see test_comparison_count below...
    user1_events.assert_has_calls(
        [
            call(EmissionInfo(signal=user1.events.id, args=(1,))),
            # call(EmissionInfo(signal=user1.events.user_name, args=("K",))),
        ]
    )
    u1_id_events.reset_mock()
    u2_id_events.reset_mock()
    user1_events.reset_mock()

    # Update user1 from user2 again, no event emission expected
    user1.update(user2)
    assert asdict(user1) == {"id": 1, "user_name": "K"}

    u1_id_events.assert_not_called()
    u2_id_events.assert_not_called()
    assert user1_events.call_count == 0


def test_update_with_inner_model_union():
    class Inner(EventedModel):
        w: str

    class AltInner(EventedModel):
        x: str

    class Outer(EventedModel):
        y: int
        z: Inner | AltInner

    original = Outer(y=1, z=Inner(w="a"))
    updated = Outer(y=2, z=AltInner(x="b"))

    original.update(updated, recurse=False)

    assert original == updated


def test_update_with_inner_model_protocol():
    @runtime_checkable
    class InnerProtocol(Protocol):
        def string(self) -> str: ...

        # Protocol fields are not successfully set without explicit validation.
        @classmethod
        def __get_validators__(cls):
            yield cls.validate

        @classmethod
        def __get_pydantic_core_schema__(cls, _source_type: Any, _handler: Any):
            from pydantic_core import core_schema

            return core_schema.no_info_plain_validator_function(cls.validate)

        @classmethod
        def validate(cls, v):
            return v

    class Inner(EventedModel):
        w: str

        def string(self) -> str:
            return self.w

    class AltInner(EventedModel):
        x: str

        def string(self) -> str:
            return self.x

    class Outer(EventedModel):
        y: int
        z: InnerProtocol

    original = Outer(y=1, z=Inner(w="a"))
    updated = Outer(y=2, z=AltInner(x="b"))

    original.update(updated, recurse=False)

    assert original == updated


def test_evented_model_signature():
    class T(EventedModel):
        x: int
        y: str = "yyy"
        z: bytes = b"zzz"

    assert isinstance(T.__signature__, inspect.Signature)
    sig = inspect.signature(T)
    assert str(sig) == "(*, x: int, y: str = 'yyy', z: bytes = b'zzz') -> None"


class MyObj:
    def __init__(self, a: int, b: str) -> None:
        self.a = a
        self.b = b

    @classmethod
    def __get_validators__(cls):
        yield cls.validate_type

    @classmethod
    def __get_pydantic_core_schema__(cls, _source_type: Any, _handler: Any):
        from pydantic_core import core_schema

        return core_schema.no_info_plain_validator_function(cls.validate_type)

    @classmethod
    def validate_type(cls, val):
        # turn a generic dict into object
        if isinstance(val, dict):
            a = val.get("a")
            b = val.get("b")
        elif isinstance(val, MyObj):
            return val
        # perform additional validation here
        return cls(a, b)

    def __eq__(self, other):
        return self.__dict__ == other.__dict__

    def _json_encode(self):
        return self.__dict__


def test_evented_model_serialization():
    class Model(EventedModel):
        """Demo evented model."""

        obj: MyObj

        @field_serializer("obj")
        def serialize_dt(self, dt: MyObj) -> dict:
            return dt.__dict__

    m = Model(obj=MyObj(1, "hi"))
    raw = asjson(m)
    if PYDANTIC_V2:
        assert raw == '{"obj":{"a":1,"b":"hi"}}'
        deserialized = Model.model_validate_json(raw)
    else:
        assert raw == '{"obj": {"a": 1, "b": "hi"}}'
        deserialized = Model.parse_raw(raw)
    assert deserialized == m


def test_nested_evented_model_serialization():
    """Test that encoders on nested sub-models can be used by top model."""

    class NestedModel(EventedModel):
        obj: MyObj

        @field_serializer("obj")
        def serialize_dt(self, dt: MyObj) -> dict:
            return dt.__dict__

    class Model(EventedModel):
        nest: NestedModel

    m = Model(nest={"obj": {"a": 1, "b": "hi"}})
    raw = asjson(m)
    if PYDANTIC_V2:
        assert raw == r'{"nest":{"obj":{"a":1,"b":"hi"}}}'
        deserialized = Model.model_validate_json(raw)
    else:
        assert raw == r'{"nest": {"obj": {"a": 1, "b": "hi"}}}'
        deserialized = Model.parse_raw(raw)
    assert deserialized == m


def test_evented_model_dask_delayed():
    """Test that evented models work with dask delayed objects"""
    dd = pytest.importorskip("dask.delayed")
    dask = pytest.importorskip("dask")

    class MyObject(EventedModel):
        attribute: dd.Delayed

        if PYDANTIC_V2:
            model_config = {"arbitrary_types_allowed": True}
        else:

            class Config:
                arbitrary_types_allowed = True

    @dask.delayed
    def my_function():
        pass

    o1 = MyObject(attribute=my_function)

    # check that equality checking works as expected
    assert o1 == o1


class T(EventedModel):
    a: int = 1
    b: int = 1

    @property
    def c(self) -> list[int]:
        return [self.a, self.b]

    @c.setter
    def c(self, val: Sequence[int]):
        self.a, self.b = val

    if PYDANTIC_V2:
        model_config = {
            "allow_property_setters": True,
            "guess_property_dependencies": True,
        }
    else:

        class Config:
            allow_property_setters = True
            guess_property_dependencies = True


def test_defaults():
    class R(EventedModel):
        x: str = "hi"

    default_r = R()

    class D(EventedModel):
        a: int = 1
        b: int = 1
        r: R = default_r

    d = D()
    assert d._defaults == {"a": 1, "b": 1, "r": default_r}

    d.update({"a": 2, "r": {"x": "asdf"}}, recurse=True)
    assert asdict(d) == {"a": 2, "b": 1, "r": {"x": "asdf"}}
    assert asdict(d) != d._defaults
    d.reset()
    assert asdict(d) == d._defaults


@pytest.mark.skipif(PYDANTIC_V2, reason="enum values seem broken on pydantic")
def test_enums_as_values():
    from enum import Enum

    class MyEnum(Enum):
        A = "value"

    class SomeModel(EventedModel):
        a: MyEnum = MyEnum.A

    m = SomeModel()
    assert asdict(m) == {"a": MyEnum.A}
    with m.enums_as_values():
        assert asdict(m) == {"a": "value"}
    assert asdict(m) == {"a": MyEnum.A}


def test_properties_with_explicit_property_dependencies():
    class MyModel(EventedModel):
        a: int = 1
        b: int = 1

        @property
        def c(self) -> list[int]:
            return [self.a, self.b]

        @c.setter
        def c(self, val: Sequence[int]) -> None:
            self.a, self.b = val

        if PYDANTIC_V2:
            model_config = {
                "allow_property_setters": True,
                "field_dependencies": {"c": ["a", "b"]},
            }
        else:

            class Config:
                allow_property_setters = True
                field_dependencies = {"c": ["a", "b"]}

    assert list(MyModel.__property_setters__) == ["c"]
    # the metaclass should have figured out that both a and b affect c
    assert MyModel.__field_dependents__ == {"a": {"c"}, "b": {"c"}}


def test_evented_model_with_property_setters():
    t = T()

    assert list(T.__property_setters__) == ["c"]
    # the metaclass should have figured out that both a and b affect c
    assert T.__field_dependents__ == {"a": {"c"}, "b": {"c"}}

    # all the fields and properties behave as expected
    assert t.c == [1, 1]
    t.a = 4
    assert t.c == [4, 1]
    t.c = [2, 3]
    assert t.c == [2, 3]
    assert t.a == 2
    assert t.b == 3


def test_evented_model_with_property_setters_events():
    t = T()
    assert "c" in t.events  # the setter has an event
    mock_a = Mock()
    mock_b = Mock()
    mock_c = Mock()
    t.events.a.connect(mock_a)
    t.events.b.connect(mock_b)
    t.events.c.connect(mock_c)

    # setting t.c emits events for all three a, b, and c
    t.c = [10, 20]
    mock_a.assert_called_with(10)
    mock_b.assert_called_with(20)
    mock_c.assert_called_with([10, 20])
    assert t.a == 10
    assert t.b == 20

    mock_a.reset_mock()
    mock_b.reset_mock()
    mock_c.reset_mock()

    # setting t.a emits events for a and c, but not b
    # this is because we declared c to be dependent on ['a', 'b']
    t.a = 5
    mock_a.assert_called_with(5)
    mock_c.assert_called_with([5, 20])
    mock_b.assert_not_called()
    assert t.c == [5, 20]
    mock_a.reset_mock()
    t.a = 5  # no change, no events
    mock_a.assert_not_called()


def test_non_setter_with_dependencies() -> None:
    with pytest.raises(
        ValueError, match=r"Fields with dependencies must be fields or property.setters"
    ):

        class M(EventedModel):
            x: int

            @property
            def y(self): ...

            @y.setter
            def y(self, v): ...

            if PYDANTIC_V2:
                model_config = {
                    "allow_property_setters": True,
                    "field_dependencies": {"a": []},
                }
            else:

                class Config:
                    allow_property_setters = True
                    field_dependencies = {"a": []}


def test_unrecognized_property_dependencies():
    with pytest.warns(UserWarning, match="cannot depend on unrecognized attribute"):

        class M(EventedModel):
            x: int

            @property
            def y(self): ...

            @y.setter
            def y(self, v): ...

            if PYDANTIC_V2:
                model_config = {
                    "allow_property_setters": True,
                    "field_dependencies": {"y": ["b"]},
                }
            else:

                class Config:
                    allow_property_setters = True
                    field_dependencies = {"y": ["b"]}


@pytest.mark.skipif(PYDANTIC_V2, reason="pydantic 2 does not support this")
def test_setattr_before_init():
    class M(EventedModel):
        _x: int = PrivateAttr()

        def __init__(_model_self_, x: int, **data) -> None:
            _model_self_._x = x
            super().__init__(**data)

        @property
        def x(self) -> int:
            return self._x

    m = M(x=2)
    assert m.x == 2


def test_setter_inheritance():
    class M(EventedModel):
        _x: int = PrivateAttr()

        def __init__(self, x: int, **data: Any) -> None:
            super().__init__(**data)
            self.x = x

        @property
        def x(self) -> int:
            return self._x

        @x.setter
        def x(self, v: int) -> None:
            self._x = v

        if PYDANTIC_V2:
            model_config = {"allow_property_setters": True}
        else:

            class Config:
                allow_property_setters = True

    assert M(x=2).x == 2

    class N(M): ...

    assert N(x=2).x == 2

    with pytest.raises(ValueError, match="Cannot set 'allow_property_setters' to"):

        class Bad(M):
            if PYDANTIC_V2:
                model_config = {"allow_property_setters": False}
            else:

                class Config:
                    allow_property_setters = False


def test_derived_events() -> None:
    class Model(EventedModel):
        a: int

        @property
        def b(self) -> int:
            return self.a + 1

        @b.setter
        def b(self, b: int) -> None:
            self.a = b - 1

        if PYDANTIC_V2:
            model_config = {
                "allow_property_setters": True,
                "field_dependencies": {"b": ["a"]},
            }
        else:

            class Config:
                allow_property_setters = True
                field_dependencies = {"b": ["a"]}

    mock_a = Mock()
    mock_b = Mock()
    m = Model(a=0)
    m.events.a.connect(mock_a)
    m.events.b.connect(mock_b)
    m.b = 3
    mock_a.assert_called_once_with(2)
    mock_b.assert_called_once_with(3)


def test_root_validator_events():
    class Model(EventedModel):
        x: int
        y: int

        if PYDANTIC_V2:
            from pydantic import model_validator

            model_config = {
                "validate_assignment": True,
                "field_dependencies": {"y": ["x"]},
            }

            @model_validator(mode="before")
            def check(cls, values: dict) -> dict:
                x = values["x"]
                values["y"] = min(values["y"], x)
                return values

        else:
            from pydantic import root_validator

            class Config:
                validate_assignment = True
                field_dependencies = {"y": ["x"]}

            @root_validator
            def check(cls, values: dict) -> dict:
                x = values["x"]
                values["y"] = min(values["y"], x)
                return values

    m = Model(x=2, y=1)
    xmock = Mock()
    ymock = Mock()
    m.events.x.connect(xmock)
    m.events.y.connect(ymock)
    m.x = 0
    assert m.y == 0
    xmock.assert_called_once_with(0)
    ymock.assert_called_once_with(0)

    xmock.reset_mock()
    ymock.reset_mock()

    m.x = 2
    assert m.y == 0
    xmock.assert_called_once_with(2)
    ymock.assert_not_called()


def test_deprecation() -> None:
    with pytest.warns(DeprecationWarning, match="Use 'field_dependencies' instead"):

        class MyModel(EventedModel):
            a: int = 1
            b: int = 1

            if PYDANTIC_V2:
                model_config = {"property_dependencies": {"a": ["b"]}}
            else:

                class Config:
                    property_dependencies = {"a": ["b"]}

        assert MyModel.__field_dependents__ == {"b": {"a"}}


def test_comparison_count() -> None:
    """Test that we only compare fields that are actually connected to events."""

    class Model(EventedModel):
        a: int

        @property
        def b(self) -> int:
            return self.a + 1

        @b.setter
        def b(self, b: int) -> None:
            self.a = b - 1

        if PYDANTIC_V2:
            model_config = {
                "allow_property_setters": True,
                "field_dependencies": {"b": ["a"]},
            }
        else:

            class Config:
                allow_property_setters = True
                field_dependencies = {"b": ["a"]}

    # pick whether to mock v1 or v2 modules
    model_module = sys.modules[type(Model).__module__]

    m = Model(a=0)
    b_mock = Mock()
    with patch.object(
        model_module,
        "_check_field_equality",
        wraps=model_module._check_field_equality,
    ) as check_mock:
        m.a = 1

    check_mock.assert_not_called()
    b_mock.assert_not_called()

    m.events.b.connect(b_mock)
    with patch.object(
        model_module,
        "_check_field_equality",
        wraps=model_module._check_field_equality,
    ) as check_mock:
        m.a = 3
    check_mock.assert_has_calls([call(Model, "a", 3, 1), call(Model, "b", 4, 2)])
    b_mock.assert_called_once_with(4)


def test_connect_only_to_events() -> None:
    """Make sure that we still make comparison and emit events when connecting
    only to the events group itself."""

    class Model(EventedModel):
        a: int

    # pick whether to mock v1 or v2 modules
    model_module = sys.modules[type(Model).__module__]

    m = Model(a=0)
    mock1 = Mock()
    with patch.object(
        model_module,
        "_check_field_equality",
        wraps=model_module._check_field_equality,
    ) as check_mock:
        m.a = 1

    check_mock.assert_not_called()
    mock1.assert_not_called()

    m.events.all.connect(mock1)
    with patch.object(
        model_module,
        "_check_field_equality",
        wraps=model_module._check_field_equality,
    ) as check_mock:
        m.a = 3
    check_mock.assert_has_calls([call(Model, "a", 3, 1)])
    mock1.assert_called_once()


def test_if_event_is_emitted_only_once() -> None:
    """Check if, for complex property setters, the event is emitted only once."""

    class SampleClass(EventedModel):
        a: int = 1
        b: int = 2

        if PYDANTIC_V2:
            model_config = {
                "allow_property_setters": True,
                "guess_property_dependencies": True,
            }
        else:

            class Config:
                allow_property_setters = True
                guess_property_dependencies = True

        @property
        def c(self):
            return self.a + self.b

        @c.setter
        def c(self, value):
            self.a = value - self.b

        @property
        def d(self):
            return self.a + self.b

        @d.setter
        def d(self, value):
            self.a = value // 2
            self.b = value - self.a

    s = SampleClass()
    a_m = Mock()
    c_m = Mock()
    d_m = Mock()
    s.events.a.connect(a_m)
    s.events.c.connect(c_m)
    s.events.d.connect(d_m)

    s.d = 5
    a_m.assert_called_once()
    c_m.assert_called_once()
    d_m.assert_called_once()


@pytest.mark.parametrize(
    "mode",
    [
        ReemissionMode.IMMEDIATE,
        ReemissionMode.QUEUED,
        {"a": ReemissionMode.IMMEDIATE, "b": ReemissionMode.QUEUED},
        {"a": ReemissionMode.IMMEDIATE, "b": "err"},
        {"a": ReemissionMode.QUEUED},
        {},
        "err",
    ],
)
def test_evented_model_reemission(mode: str | dict) -> None:
    err = mode == "err" or (isinstance(mode, dict) and "err" in mode.values())
    with (
        pytest.raises(ValueError, match="Invalid reemission") if err else nullcontext()
    ):

        class Model(EventedModel):
            a: int
            b: int

            if PYDANTIC_V2:
                model_config = {"reemission": mode}
            else:

                class Config:
                    reemission = mode

    if err:
        return

    m = Model(a=1, b=2)
    if isinstance(mode, dict):
        assert m.events.a._reemission == mode.get("a", ReemissionMode.LATEST)
        assert m.events.b._reemission == mode.get("b", ReemissionMode.LATEST)
    else:
        assert m.events.a._reemission == mode
        assert m.events.b._reemission == mode


@pytest.mark.skipif(not PYDANTIC_V2, reason="computed_field added in v2")
def test_computed_field() -> None:
    from pydantic import computed_field

    class MyModel(EventedModel):
        a: int = 1
        b: int = 1

        @computed_field
        @property
        def c(self) -> list[int]:
            return [self.a, self.b]

        @c.setter
        def c(self, val: Sequence[int]) -> None:
            self.a, self.b = val

        model_config = {
            "allow_property_setters": True,
            "field_dependencies": {"c": ["a", "b"]},  # type: ignore [typeddict-unknown-key]
        }

    mock_a = Mock()
    mock_b = Mock()
    mock_c = Mock()
    m = MyModel()
    m.events.a.connect(mock_a)
    m.events.b.connect(mock_b)
    m.events.c.connect(mock_c)

    m.c = [10, 20]
    mock_a.assert_called_with(10)
    mock_b.assert_called_with(20)
    mock_c.assert_called_with([10, 20])

    mock_a.reset_mock()
    mock_c.reset_mock()

    m.a = 5
    mock_a.assert_called_with(5)
    mock_c.assert_called_with([5, 20])


@pytest.mark.skipif(not PYDANTIC_V2, reason="computed_field added in v2")
def test_private_field_dependents():
    from pydantic import PrivateAttr, computed_field

    from psygnal import EventedModel

    class MyModel(EventedModel):
        _items_dict: dict[str, int] = PrivateAttr(default_factory=dict)

        @computed_field  # type: ignore [prop-decorator]
        @property
        def item_names(self) -> list[str]:
            return list(self._items_dict)

        @computed_field  # type: ignore [prop-decorator]
        @property
        def item_sum(self) -> int:
            return sum(self._items_dict.values())

        def add_item(self, name: str, value: int) -> None:
            if name in self._items_dict:
                raise ValueError(f"Name {name} already exists!")
            self._items_dict = {**self._items_dict, name: value}

        # Ideally the following would work
        model_config = {
            "field_dependencies": {  # type: ignore [typeddict-unknown-key]
                "item_names": ["_items_dict"],
                "item_sum": ["_items_dict"],
            }
        }

    m = MyModel()
    item_sum_mock = Mock()
    item_names_mock = Mock()
    m.events.item_sum.connect(item_sum_mock)
    m.events.item_names.connect(item_names_mock)
    m.add_item("a", 1)
    item_sum_mock.assert_called_once_with(1)
    item_names_mock.assert_called_once_with(["a"])
    item_sum_mock.reset_mock()
    item_names_mock.reset_mock()
    m.add_item("b", 2)
    item_sum_mock.assert_called_once_with(3)
    item_names_mock.assert_called_once_with(["a", "b"])

    # check direct access ass well
    item_sum_mock.reset_mock()
    m._items_dict = m._items_dict.copy()
    assert item_sum_mock.call_count == 0
    m._items_dict = {}
    item_sum_mock.assert_called_once_with(0)


def test_primary_vs_dependent_optimization() -> None:
    """Test that dependent properties are not checked when no primary changes occur.

    This test verifies the optimization where, if no primary fields actually change
    values, the system skips checking dependent properties entirely.

    The current implementation processes primary changes first, and if none of them
    result in actual value changes, it skips checking dependent properties entirely.
    """

    class Model(EventedModel):
        a: int = 1

        @property
        def b(self) -> int:
            return self.a + 1

        @b.setter
        def b(self, b: int) -> None:
            self.a = b - 1

        if PYDANTIC_V2:
            model_config = {
                "allow_property_setters": True,
                "field_dependencies": {"b": ["a"]},
            }
        else:

            class Config:
                allow_property_setters = True
                field_dependencies = {"b": ["a"]}

    # pick whether to mock v1 or v2 modules
    model_module = sys.modules[type(Model).__module__]

    m = Model(a=5)

    # Connect listeners to both primary field and dependent property
    # so that the system has reason to track changes
    mock_a = Mock()
    mock_b = Mock()
    m.events.a.connect(mock_a)
    m.events.b.connect(mock_b)

    # Set field 'a' to the same value it already has (no actual change)
    with patch.object(
        model_module,
        "_check_field_equality",
        wraps=model_module._check_field_equality,
    ) as check_mock:
        m.a = 5  # Same value, no change

    # implementation should only check the primary field 'a'
    # and skip checking dependent property 'b' since no primary change occurred
    check_mock.assert_has_calls([call(Model, "a", 5, 5)])
    assert check_mock.call_count == 1, (
        f"Expected 1 equality check, got {check_mock.call_count}"
    )

    # No events should be emitted since no actual changes occurred
    mock_a.assert_not_called()
    mock_b.assert_not_called()


@pytest.mark.skipif(not PYDANTIC_V2, reason="v2 serialization features")
def test_serialization_and_schema():
    class TestModel(EventedModel):
        name: str
        elist_of_str: EventedList[str] = Field(default_factory=EventedList)
        eset_of_str: EventedSet[str] = Field(default_factory=EventedSet)
        edict_of_str: EventedDict[str, str] = Field(default_factory=EventedDict)

    model = TestModel(name="Test")
    assert isinstance(model.elist_of_str, EventedList)
    assert isinstance(model.eset_of_str, EventedSet)
    assert isinstance(model.edict_of_str, EventedDict)

    dumped = model.model_dump(mode="python")
    assert isinstance(dumped["elist_of_str"], EventedList)
    assert isinstance(dumped["eset_of_str"], EventedSet)
    assert isinstance(dumped["edict_of_str"], EventedDict)

    dumped = model.model_dump(mode="json")
    assert isinstance(dumped["elist_of_str"], list)
    assert isinstance(dumped["eset_of_str"], list)
    assert isinstance(dumped["edict_of_str"], dict)

    json_dump = model.model_dump_json()
    assert isinstance(json_dump, str)
    assert TestModel.model_json_schema(mode="serialization")
    assert TestModel.model_json_schema(mode="validation")



import importlib.util
import os
import subprocess
import warnings
from pathlib import Path

import pytest

import psygnal


def test_hook_content():
    spec = importlib.util.spec_from_file_location(
        "hook",
        os.path.join(
            os.path.dirname(psygnal.__file__), "_pyinstaller_util", "hook-psygnal.py"
        ),
    )
    hook = importlib.util.module_from_spec(spec)
    spec.loader.exec_module(hook)

    assert "mypy_extensions" in hook.hiddenimports

    if not psygnal._compiled:
        return

    assert "psygnal._dataclass_utils" in hook.hiddenimports


@pytest.mark.skipif(not os.getenv("CI"), reason="slow test")
def test_pyintstaller_hiddenimports(tmp_path: Path) -> None:
    with warnings.catch_warnings():
        warnings.simplefilter("ignore")
        pyi_main = pytest.importorskip("PyInstaller.__main__")

    build_path = tmp_path / "build"
    dist_path = tmp_path / "dist"
    app_name = "psygnal_test"
    app = tmp_path / f"{app_name}.py"
    app.write_text("\n".join(["import psygnal", "print(psygnal.__version__)"]))

    args = [
        # Place all generated files in ``tmp_path``.
        "--workpath",
        str(build_path),
        "--distpath",
        str(dist_path),
        "--specpath",
        str(tmp_path),
        str(app),
    ]

    with warnings.catch_warnings():
        warnings.simplefilter("ignore")
        # silence warnings about deprecations
        pyi_main.run(args)
    subprocess.run([str(dist_path / app_name / app_name)], check=True)



import re

import pytest

import psygnal.testing as pt
from psygnal import Signal


class MyObject:
    changed = Signal()
    value_changed = Signal(int)


def test_assert_emitted() -> None:
    obj = MyObject()

    with pt.assert_emitted(obj.changed) as tester:
        obj.changed.emit()

    assert isinstance(tester, pt.SignalTester)

    with pytest.raises(
        AssertionError, match=r"Expected 'changed' to have been emitted."
    ):
        with pt.assert_emitted(obj.changed):
            pass


def test_assert_emitted_once():
    obj = MyObject()
    with pt.assert_emitted_once(obj.changed) as tester:
        obj.changed.emit()
    assert isinstance(tester, pt.SignalTester)

    with pytest.raises(
        AssertionError,
        match=r"Expected 'changed' to have been emitted once. Emitted 2 times.",
    ):
        with pt.assert_emitted_once(obj.changed):
            obj.changed.emit()
            obj.changed.emit()


def test_assert_not_emitted() -> None:
    obj = MyObject()
    with pt.assert_not_emitted(obj.changed) as tester:
        pass

    assert isinstance(tester, pt.SignalTester)

    with pytest.raises(
        AssertionError,
        match=r"Expected 'changed' to not have been emitted. Emitted once.",
    ):
        with pt.assert_not_emitted(obj.changed):
            obj.changed.emit()

    with pytest.raises(
        AssertionError,
        match=r"Expected 'changed' to not have been emitted. Emitted 4 times.",
    ):
        with pt.assert_not_emitted(obj.changed):
            obj.changed.emit()
            obj.changed.emit()
            obj.changed.emit()
            obj.changed.emit()


def test_assert_emitted_with() -> None:
    obj = MyObject()
    with pt.assert_emitted_with(obj.value_changed, 42) as tester:
        obj.value_changed.emit(41)
        obj.value_changed.emit(42)

    assert isinstance(tester, pt.SignalTester)

    with pytest.raises(
        AssertionError,
        match=re.escape(
            "Expected 'value_changed' to have been emitted with arguments (42,)."
            "\nActual: not emitted"
        ),
    ):
        with pt.assert_emitted_with(obj.value_changed, 42):
            pass

    with pytest.raises(
        AssertionError,
        match=re.escape(
            "Expected 'value_changed' to have been emitted with arguments (42,)."
            "\nActual: (43,)"
        ),
    ):
        with pt.assert_emitted_with(obj.value_changed, 42):
            obj.value_changed.emit(42)
            obj.value_changed.emit(43)


def test_assert_emitted_once_with() -> None:
    obj = MyObject()
    with pt.assert_emitted_once_with(obj.value_changed, 42) as tester:
        obj.value_changed.emit(42)

    assert isinstance(tester, pt.SignalTester)

    with pytest.raises(
        AssertionError,
        match=re.escape(
            "Expected 'value_changed' to have been emitted exactly once. "
            "Emitted 2 times."
        ),
    ):
        with pt.assert_emitted_once_with(obj.value_changed, 42):
            obj.value_changed.emit(42)
            obj.value_changed.emit(42)

    with pytest.raises(
        AssertionError,
        match=re.escape(
            "Expected 'value_changed' to have been emitted once with arguments (42,)."
            "\nActual: (43,)"
        ),
    ):
        with pt.assert_emitted_once_with(obj.value_changed, 42):
            obj.value_changed.emit(43)


def test_assert_ever_emitted_with() -> None:
    obj = MyObject()

    with pt.assert_ever_emitted_with(obj.value_changed, 42) as tester:
        obj.value_changed.emit(41)
        obj.value_changed.emit(42)
        obj.value_changed.emit(43)

    assert isinstance(tester, pt.SignalTester)

    with pytest.raises(
        AssertionError,
        match=re.escape(
            "Expected 'value_changed' to have been emitted at least once with "
            "arguments (42,)."
            "\nActual: not emitted"
        ),
    ):
        with pt.assert_ever_emitted_with(obj.value_changed, 42):
            pass

    with pytest.raises(
        AssertionError,
        match=re.escape(
            "Expected 'value_changed' to have been emitted at least once with "
            "arguments (42,)."
            "\nActual: (43,)"
        ),
    ):
        with pt.assert_ever_emitted_with(obj.value_changed, 42):
            obj.value_changed.emit(43)

    with pytest.raises(
        AssertionError,
        match=re.escape(
            "Expected 'value_changed' to have been emitted at least once with "
            "arguments (42,)."
            "\nActual: [(41,), (42, 43)]"
        ),
    ):
        with pt.assert_ever_emitted_with(obj.value_changed, 42):
            obj.value_changed.emit(41)
            obj.value_changed.emit(42, 43)


def test_signal_tester() -> None:
    obj = MyObject()
    tester = pt.SignalTester(obj.changed)
    tester.connect()
    assert tester.signal_name == "changed"
    assert tester.mock.call_count == 0

    obj.changed.emit()

    tester.assert_emitted_once()
    tester.assert_emitted()
    tester.assert_emitted_with()
    assert tester.emit_count == 1
    tester.reset()
    assert tester.emit_count == 0

    tester2 = pt.SignalTester(obj.value_changed)

    with tester2:
        obj.value_changed.emit(42)
        obj.value_changed.emit(43)

    tester2.assert_emitted()
    tester2.assert_emitted_with(43)
    tester2.assert_ever_emitted_with(42)

    assert tester2.emit_args_list == [(42,), (43,)]
    assert tester2.emit_count == 2
    assert tester2.emit_args == (43,)



import operator
from dataclasses import dataclass, field
from typing import TYPE_CHECKING, ClassVar, cast, no_type_check
from unittest.mock import Mock

import numpy as np
import pytest

from psygnal import (
    EmissionInfo,
    PathStep,
    Signal,
    SignalGroup,
    SignalGroupDescriptor,
    SignalInstance,
    evented,
    get_evented_namespace,
    is_evented,
    testing,
)
from psygnal._group import SignalRelay

try:
    import pydantic.version

    PYDANTIC_V2 = pydantic.version.VERSION.startswith("2")
except ImportError:
    PYDANTIC_V2 = False


decorated_or_descriptor = pytest.mark.parametrize(
    "decorator", [True, False], ids=["decorator", "descriptor"]
)


@no_type_check
def _check_events(cls, events_ns="events"):
    obj = cls(bar=1, baz="2", qux=np.zeros(3))
    assert is_evented(obj)
    assert is_evented(cls)
    assert get_evented_namespace(cls) == events_ns
    assert isinstance(getattr(cls, events_ns), SignalGroupDescriptor)

    events = getattr(obj, events_ns)
    assert isinstance(events, SignalGroup)
    assert set(events) == {"bar", "baz", "qux"}

    mock = Mock()
    events.bar.connect(mock)
    assert obj.bar == 1
    obj.bar = 2
    assert obj.bar == 2
    mock.assert_called_once_with(2, 1)

    mock.reset_mock()
    obj.baz = "3"
    mock.assert_not_called()

    mock.reset_mock()
    events.qux.connect(mock)
    obj.qux = np.ones(3)
    mock.assert_called_once()
    assert np.array_equal(obj.qux, np.ones(3))


@decorated_or_descriptor
@pytest.mark.parametrize("kwargs", ({"slots": True}, {"slots": False}))
def test_native_dataclass(decorator: bool, kwargs: dict) -> None:
    @dataclass(**kwargs)
    class Base:
        bar: int
        baz: str
        qux: np.ndarray

    if decorator:

        @evented(equality_operators={"qux": operator.eq})  # just for test coverage
        class Foo(Base): ...

    else:

        class Foo(Base):  # type: ignore [no-redef]
            events: ClassVar[SignalGroupDescriptor] = SignalGroupDescriptor(
                equality_operators={"qux": operator.eq}
            )

    _check_events(Foo)


@decorated_or_descriptor
@pytest.mark.parametrize("slots", [True, False])
def test_attrs_dataclass(decorator: bool, slots: bool) -> None:
    from attrs import define

    @define(slots=slots)  # type: ignore [misc]
    class Base:
        bar: int
        baz: str
        qux: np.ndarray

    if decorator:

        @evented
        class Foo(Base): ...

    else:

        class Foo(Base):  # type: ignore [no-redef]
            events: ClassVar[SignalGroupDescriptor] = SignalGroupDescriptor()

    _check_events(Foo)


if PYDANTIC_V2:
    Config = {"arbitrary_types_allowed": True}
else:

    class Config:
        arbitrary_types_allowed = True


@decorated_or_descriptor
def test_pydantic_dataclass(decorator: bool) -> None:
    pytest.importorskip("pydantic")
    from pydantic.dataclasses import dataclass

    @dataclass(config=Config)
    class Base:
        bar: int
        baz: str
        qux: np.ndarray

    if decorator:

        @evented
        class Foo(Base): ...

    else:

        class Foo(Base):  # type: ignore [no-redef]
            events: ClassVar[SignalGroupDescriptor] = SignalGroupDescriptor()

    _check_events(Foo)


@decorated_or_descriptor
def test_pydantic_base_model(decorator: bool) -> None:
    pytest.importorskip("pydantic")
    from pydantic import BaseModel

    class Base(BaseModel):
        bar: int
        baz: str
        qux: np.ndarray

        if PYDANTIC_V2:
            model_config = Config
        else:
            Config = Config  # type: ignore

    if decorator:

        @evented(events_namespace="my_events")
        class Foo(Base): ...

    else:

        class Foo(Base):  # type: ignore [no-redef]
            my_events: ClassVar[SignalGroupDescriptor] = SignalGroupDescriptor()

    _check_events(Foo, "my_events")


@pytest.mark.parametrize("decorator", [True, False], ids=["decorator", "descriptor"])
def test_msgspec_struct(decorator: bool) -> None:
    if TYPE_CHECKING:
        import msgspec
    else:
        msgspec = pytest.importorskip("msgspec")  # remove when py37 is dropped

    if decorator:

        @evented
        class Foo(msgspec.Struct):
            bar: int
            baz: str
            qux: np.ndarray

    else:

        class Foo(msgspec.Struct):  # type: ignore [no-redef]
            bar: int
            baz: str
            qux: np.ndarray
            events: ClassVar[SignalGroupDescriptor] = SignalGroupDescriptor()

    _check_events(Foo)


def test_no_signals_warn() -> None:
    with pytest.warns(UserWarning, match="No mutable fields found on class"):

        @evented
        class Foo: ...

        _ = Foo().events  # type: ignore

    with pytest.warns(UserWarning, match="No mutable fields found on class"):

        class Foo2:
            events = SignalGroupDescriptor()

        _ = Foo2().events

    @dataclass
    class Foo3:
        events = SignalGroupDescriptor(warn_on_no_fields=False)

    # no warning
    _ = Foo3().events


@dataclass
class FooPicklable:
    bar: int
    events: ClassVar[SignalGroupDescriptor] = SignalGroupDescriptor(
        cache_on_instance=False
    )


def test_pickle() -> None:
    """Make sure that evented classes are still picklable."""
    import pickle

    obj = FooPicklable(1)
    obj2 = pickle.loads(pickle.dumps(obj))
    assert obj2.bar == 1


def test_get_namespace() -> None:
    @evented(events_namespace="my_events")
    @dataclass
    class Foo:
        x: int

    assert get_evented_namespace(Foo) == "my_events"
    assert is_evented(Foo)


def test_name_conflicts() -> None:
    # https://github.com/pyapp-kit/psygnal/pull/269
    from dataclasses import field

    @evented
    @dataclass
    class Foo:
        name: str
        all: bool = False
        is_uniform: bool = True
        signals: list = field(default_factory=list)

    obj = Foo("foo")
    assert obj.name == "foo"
    with pytest.warns(
        UserWarning, match=r"Names \['all', 'is_uniform', 'signals'\] are reserved"
    ):
        group = obj.events

    assert isinstance(group, SignalGroup)

    assert "name" in group
    assert isinstance(group.name, SignalInstance)
    assert group["name"] is group.name

    assert "is_uniform" in group and isinstance(group["is_uniform"], SignalInstance)
    assert "signals" in group and isinstance(group["signals"], SignalInstance)

    # group.all is always a relay
    assert isinstance(group.all, SignalRelay)

    # getitem returns the signal
    assert "all" in group and isinstance(group["all"], SignalInstance)
    assert not isinstance(group["all"], SignalRelay)

    with pytest.raises(AttributeError):  # it's not writeable
        group.all = SignalRelay({})  # type: ignore

    assert group.psygnals_uniform() is False

    @evented
    @dataclass
    class Foo2:
        psygnals_uniform: bool = True

    obj2 = Foo2()
    with pytest.warns(match=r"Name \['psygnals_uniform'\] is reserved"):
        _ = obj2.events

    @dataclass
    class Foo3:
        field: int = 1
        _psygnal_signals: str = "signals"

    with pytest.raises(
        TypeError, match="Fields on an evented class cannot start with '_psygnal'"
    ):
        _ = evented(Foo3)


def test_nesting() -> None:
    from dataclasses import dataclass, field

    @evented
    @dataclass
    class Foo:
        x: int = 1

    @evented
    @dataclass
    class Bar:
        y: int = 2
        foo: Foo = field(default_factory=Foo)

    # @evented(connect_child_events=True)  # could also use this syntax
    @dataclass
    class Baz:
        events: ClassVar[SignalGroupDescriptor] = SignalGroupDescriptor(
            connect_child_events=True
        )
        z: int = 3
        bar: Bar = field(default_factory=Bar)

    baz = Baz()
    mock = Mock()
    events: SignalGroup = baz.events
    events.all.connect(mock)

    baz.bar.foo.x = 3  # trigger nested event

    # what we expect
    expected = EmissionInfo(
        baz.bar.foo.events.x,
        (3, 1),
        path=(
            PathStep(attr="bar"),  # Baz → bar
            PathStep(attr="foo"),  # bar  → foo
            PathStep(attr="x"),  # foo  → x   (added by SignalRelay inside Foo)
        ),
    )

    mock.assert_called_with(expected)


def test_signal_relay_partial():
    """Test hash and eq methods on _relay_partial objects"""

    class T(SignalGroup):
        sig = Signal(int)

    t = T()
    a = set()
    a.add(t.all._relay_partial(PathStep(attr="some_name")))
    a.add(t.all._relay_partial(PathStep(attr="some_name")))
    assert len(a) == 1

    assert t.all._relay_partial(PathStep(attr="some_name")) in a


def test_evented_object_replacement_disconnects_old_connections():
    """Test that replacing evented objects properly disconnects the old one."""

    @evented
    @dataclass
    class A:
        x: int = 1

    @dataclass
    class M:
        events: ClassVar[SignalGroupDescriptor] = SignalGroupDescriptor(
            connect_child_events=True
        )
        d: A = field(default_factory=A)

    m = M()

    # Connect to the main events
    main_mock = Mock()
    m.events.connect(main_mock)

    # Get references to the original and new evented objects
    original_d = m.d
    new_d = A(x=99)

    # Connect directly to the original object's events to verify disconnection
    original_mock = Mock()
    original_d.events.connect(original_mock)

    # Replace the evented object
    m.d = new_d

    # Verify the replacement was detected by the main events
    assert main_mock.call_count == 1
    replacement_info = cast("EmissionInfo", main_mock.call_args[0][0])
    assert replacement_info.args == (new_d, original_d)
    assert replacement_info.path == (PathStep(attr="d"),)

    # Now modify the NEW object - should trigger events through the parent
    main_mock.reset_mock()
    new_d.x = 42

    # The main events should receive the nested change
    assert main_mock.call_count == 1
    nested_info = cast("EmissionInfo", main_mock.call_args[0][0])
    assert nested_info.args == (42, 99)
    assert nested_info.path == (PathStep(attr="d"), PathStep(attr="x"))

    # Now modify the OLD object - should NOT trigger events through the parent
    # because it should have been disconnected
    main_mock.reset_mock()
    original_d.x = 123

    # The original object's direct listeners should still work
    assert original_mock.call_count == 1

    # But the main events should NOT have been triggered (disconnected)
    assert main_mock.call_count == 0


def test_lazy_child_connection() -> None:
    """Test that child events are only connected when parent is first connected to."""

    @evented
    @dataclass
    class Child:
        y: int = 2

    @evented
    @dataclass
    class Parent:
        child: Child
        x: int = 1

    # Create parent with child
    child = Child()
    parent = Parent(child=child)

    # Before any connections, neither should have relay slots
    assert len(parent.events.all) == 0
    assert len(child.events.all) == 0

    mock = Mock()
    parent.events.all.connect(mock)

    # After connection
    # parent should have our listener, child should have relay connection
    assert len(parent.events.all) == 1
    assert len(child.events.all) == 1

    # Test that child events propagate to parent
    child.y = 10

    mock.assert_called_once_with(
        EmissionInfo(
            child.events.y, (10, 2), path=(PathStep(attr="child"), PathStep(attr="y"))
        )
    )


def test_team_example() -> None:
    @evented
    @dataclass
    class Person:
        name: str = ""
        age: int = 0

    @evented
    @dataclass
    class Team:
        name: str = ""
        leader: Person = field(default_factory=Person)

    team = Team()

    # This will trigger the listener above
    team_level_info = EmissionInfo(
        team.leader.events.name,
        ("Alice", ""),
        path=(PathStep(attr="leader"), PathStep(attr="name")),
    )
    team_leader_level_info = EmissionInfo(
        team.leader.events.name, ("Alice", ""), path=(PathStep(attr="name"),)
    )
    with (
        testing.assert_emitted_once_with(team.events.all, team_level_info),
        testing.assert_emitted_once_with(
            team.leader.events.all, team_leader_level_info
        ),
        testing.assert_emitted_once_with(team.leader.events.name, "Alice", ""),
        testing.assert_not_emitted(team.events.leader),
    ):
        team.leader.name = "Alice"


def test_signal_instance_emits_on_subevents() -> None:
    @evented
    @dataclass
    class Person:
        name: str = ""
        age: int = 0

    @evented
    @dataclass
    class Team:
        name: str = ""
        leader: Person = field(default_factory=Person)

    team = Team(name="A-Team", leader=Person(name="Hannibal", age=59))

    mock = Mock()
    team.events.leader.connect(mock, emit_on_evented_child_events=True)
    team.leader.age = 60
    mock.assert_called_once_with(Person(name="Hannibal", age=60), None)



from unittest.mock import Mock

from psygnal.containers import Selection


def test_add_and_remove_from_selection():
    selection = Selection()
    selection.events._current = Mock()
    assert not selection._current
    assert not selection
    selection.add(1)
    selection._current = 1
    selection.events._current.emit.assert_called_once()

    assert 1 in selection
    assert selection._current == 1

    selection.remove(1)
    assert not selection


def test_update_active_called_on_selection_change():
    selection = Selection()
    selection._update_active = Mock()
    selection.add(1)
    selection._update_active.assert_called_once()


def test_active_event_emitted_on_selection_change():
    selection = Selection()
    selection.events.active = Mock()
    assert not selection.active
    selection.add(1)
    assert selection.active == 1
    selection.events.active.emit.assert_called_once()


def test_current_setter():
    """Current event should only emit if value changes."""
    selection = Selection()
    selection._current = 1
    selection.events._current = Mock()
    selection._current = 1
    selection.events._current.emit.assert_not_called()
    selection._current = 2
    selection.events._current.emit.assert_called_once()


def test_active_setter():
    """Active setter should make value the only selected item, make it current and
    emit the active event."""
    selection = Selection()
    selection.events.active = Mock()
    assert not selection._current
    selection.active = 1
    assert selection.active == 1
    assert selection._current == 1
    selection.events.active.emit.assert_called_once()


def test_select_only():
    mock = Mock()
    selection = Selection([1, 2])
    selection.active = 1
    assert selection.active == 1
    selection.events.items_changed.connect(mock)
    selection.select_only(2)
    mock.assert_called_once_with((2,), (1,))
    assert selection.active == 2


def test_clear():
    selection = Selection([1, 2])
    selection._current = 2
    assert len(selection) == 2
    selection.clear(keep_current=True)
    assert len(selection) == 0
    assert selection._current == 2
    selection.clear(keep_current=False)
    assert selection._current is None


def test_toggle():
    selection = Selection()
    selection.symmetric_difference_update = Mock()
    selection.toggle(1)
    selection.symmetric_difference_update.assert_called_once()


def test_emit_change():
    """emit change is overridden to also update the active value."""
    selection = Selection()
    selection._update_active = Mock()
    selection._emit_change((None,), (None,))
    selection._update_active.assert_called_once()


def test_hash():
    assert hash(Selection())


def test_replace_selection():
    mock = Mock()
    selection = Selection([1, 2, 3])

    selection.events.items_changed.connect(mock)
    selection.replace_selection([3, 4, 5])
    mock.assert_called_once_with((4, 5), (1, 2))
    assert set(selection) == {3, 4, 5}



import os
from copy import copy
from typing import cast
from unittest.mock import Mock, call

import numpy as np
import pytest

from psygnal import EmissionInfo, PathStep, Signal, SignalGroup
from psygnal.containers import EventedList


@pytest.fixture
def regular_list():
    return list(range(5))


@pytest.fixture
def test_list(regular_list):
    test_list = EventedList(regular_list)
    test_list.events = Mock(wraps=test_list.events)
    return test_list


@pytest.mark.parametrize(
    "meth",
    [
        # METHOD, ARGS, EXPECTED EVENTS
        # primary interface
        ("insert", (2, 10), ("inserting", "inserted")),  # create
        ("__getitem__", (2,), ()),  # read
        ("__setitem__", (2, 3), ("changed",)),  # update
        ("__setitem__", (slice(2), [1, 2]), ("changed",)),  # update slice
        ("__setitem__", (slice(2, 2), [1, 2]), ("changed",)),  # update slice
        ("__delitem__", (2,), ("removing", "removed")),  # delete
        (
            "__delitem__",
            (slice(2),),
            ("removing", "removed") * 2,
        ),
        ("__delitem__", (slice(0, 0),), ("removing", "removed")),
        (
            "__delitem__",
            (slice(-3),),
            ("removing", "removed") * 2,
        ),
        (
            "__delitem__",
            (slice(-2, None),),
            ("removing", "removed") * 2,
        ),
        # inherited interface
        ("append", (3,), ("inserting", "inserted")),
        ("clear", (), ("removing", "removed") * 5),
        ("count", (3,), ()),
        ("extend", ([7, 8, 9],), ("inserting", "inserted") * 3),
        ("index", (3,), ()),
        ("pop", (-2,), ("removing", "removed")),
        ("remove", (3,), ("removing", "removed")),
        ("reverse", (), ("reordered",)),
        ("__add__", ([7, 8, 9],), ()),
        ("__iadd__", ([7, 9],), ("inserting", "inserted") * 2),
        ("__radd__", ([7, 9],), ("inserting", "inserted") * 2),
        # sort?
    ],
    ids=lambda x: x[0],
)
def test_list_interface_parity(test_list, regular_list, meth):
    test_list.events = cast("Mock", test_list.events)

    method_name, args, expected = meth
    test_list_method = getattr(test_list, method_name)
    assert tuple(test_list) == tuple(regular_list)
    if hasattr(regular_list, method_name):
        regular_list_method = getattr(regular_list, method_name)
        assert test_list_method(*args) == regular_list_method(*args)
        assert tuple(test_list) == tuple(regular_list)
    else:
        test_list_method(*args)  # smoke test

    for c, expect in zip(test_list.events.call_args_list, expected, strict=False):
        event = c.args[0]
        assert event.type == expect


def test_delete(test_list):
    assert test_list == [0, 1, 2, 3, 4]

    del test_list[1]
    assert test_list == [0, 2, 3, 4]

    del test_list[2:]
    assert test_list == [0, 2]


@pytest.mark.xfail("i686" in os.getenv("AUDITWHEEL_PLAT", ""), reason="failing on i686")
def test_hash(test_list):
    assert id(test_list) == hash(test_list)

    b = EventedList([2, 3], hashable=False)
    with pytest.raises(TypeError):
        hash(b)


def test_repr(test_list):
    assert repr(test_list) == "EventedList([0, 1, 2, 3, 4])"


def test_reverse(test_list):
    assert test_list == [0, 1, 2, 3, 4]
    test_list.reverse()
    test_list.events.reordered.emit.assert_called_once()
    test_list.events.changed.emit.assert_not_called()
    assert test_list == [4, 3, 2, 1, 0]

    test_list.events.reordered.emit.reset_mock()
    test_list.reverse(emit_individual_events=True)
    test_list.events.reordered.emit.assert_called_once()
    assert test_list.events.changed.emit.call_count == 4
    test_list.events.changed.emit.assert_has_calls(
        [call(0, 4, 0), call(4, 0, 4), call(1, 3, 1), call(3, 1, 3)]
    )
    assert test_list == [0, 1, 2, 3, 4]


def test_list_interface_exceptions(test_list):
    bad_index = {"a": "dict"}
    with pytest.raises(TypeError):
        test_list[bad_index]

    with pytest.raises(TypeError):
        test_list[bad_index] = 1

    with pytest.raises(TypeError):
        del test_list[bad_index]

    with pytest.raises(TypeError):
        test_list.insert([bad_index], 0)


def test_copy(test_list, regular_list):
    """Copying an evented list should return a same-class evented list."""
    new_test = test_list.copy()
    new_reg = regular_list.copy()
    assert id(new_test) != id(test_list)
    assert new_test == test_list
    assert tuple(new_test) == tuple(test_list) == tuple(new_reg)
    test_list.events.assert_not_called()


def test_array_like_setitem():
    """Test that EventedList.__setitem__ works for array-like items"""
    array = np.array((10, 10))
    evented_list = EventedList([array])
    evented_list[0] = array


def test_slice(test_list, regular_list):
    """Slicing an evented list should return a same-class evented list."""
    test_slice = test_list[1:3]
    regular_slice = regular_list[1:3]
    assert tuple(test_slice) == tuple(regular_slice)
    assert isinstance(test_slice, test_list.__class__)

    change_emit = test_list.events.changed.emit

    assert test_list == [0, 1, 2, 3, 4]
    test_list[1:3] = [6, 7, 8]
    assert test_list == [0, 6, 7, 8, 3, 4]
    change_emit.assert_called_with(slice(1, 3, None), [1, 2], [6, 7, 8])

    with pytest.raises(ValueError) as e:
        test_list[1:6:2] = [6, 7, 8, 6, 7]
    assert str(e.value).startswith("attempt to assign sequence of size 5 to extended ")

    test_list[1:6:2] = [9, 9, 9]
    assert test_list == [0, 9, 7, 9, 3, 9]
    change_emit.assert_called_with(slice(1, 6, 2), [6, 8, 4], [9, 9, 9])

    with pytest.raises(TypeError) as e2:
        test_list[1:3] = 1
    assert str(e2.value) == "Can only assign an iterable to slice"


def test_move(test_list: EventedList) -> None:
    """Test the that we can move objects with the move method"""
    test_list.events = cast("Mock", test_list.events)

    def _fail() -> None:
        raise AssertionError("unexpected event called")

    test_list.events.removing.connect(_fail)
    test_list.events.removed.connect(_fail)
    test_list.events.inserting.connect(_fail)
    test_list.events.inserted.connect(_fail)

    before = list(test_list)
    assert before == [0, 1, 2, 3, 4]  # from fixture
    # pop the object at 0 and insert at current position 3
    test_list.move(0, 3)
    expectation = [1, 2, 0, 3, 4]
    assert test_list != before
    assert test_list == expectation
    test_list.events.moving.emit.assert_called_once_with(0, 3)
    test_list.events.moved.emit.assert_called_once_with(0, 2, 0)
    test_list.events.reordered.emit.assert_called_once()

    test_list.events.moving.emit.reset_mock()
    test_list.move(2, 2)
    test_list.events.moving.emit.assert_not_called()  # noop

    # move the other way
    # pop the object at 3 and insert at current position 0
    assert test_list == [1, 2, 0, 3, 4]
    test_list.move(3, 0)
    assert test_list == [3, 1, 2, 0, 4]

    # negative index destination
    test_list.move(1, -2)
    assert test_list == [3, 2, 0, 1, 4]


BASIC_INDICES: list[tuple] = [
    ((2,), 0, [2, 0, 1, 3, 4, 5, 6, 7]),  # move single item
    ([0, 2, 3], 6, [1, 4, 5, 0, 2, 3, 6, 7]),  # move back
    ([4, 7], 1, [0, 4, 7, 1, 2, 3, 5, 6]),  # move forward
    ([0, 5, 6], 3, [1, 2, 0, 5, 6, 3, 4, 7]),  # move in between
    ([1, 3, 5, 7], 3, [0, 2, 1, 3, 5, 7, 4, 6]),  # same as above
    ([0, 2, 3, 2, 3], 6, [1, 4, 5, 0, 2, 3, 6, 7]),  # strip dupe indices
]
OTHER_INDICES: list[tuple] = [
    ([7, 4], 1, [0, 7, 4, 1, 2, 3, 5, 6]),  # move forward reorder
    ([3, 0, 2], 6, [1, 4, 5, 3, 0, 2, 6, 7]),  # move back reorder
    ((2, 4), -2, [0, 1, 3, 5, 6, 2, 4, 7]),  # negative indexing
    ([slice(None, 3)], 6, [3, 4, 5, 0, 1, 2, 6, 7]),  # move slice back
    ([slice(5, 8)], 2, [0, 1, 5, 6, 7, 2, 3, 4]),  # move slice forward
    ([slice(1, 8, 2)], 3, [0, 2, 1, 3, 5, 7, 4, 6]),  # move slice between
    ([slice(None, 8, 3)], 4, [1, 2, 0, 3, 6, 4, 5, 7]),
    ([slice(None, 8, 3), 0, 3, 6], 4, [1, 2, 0, 3, 6, 4, 5, 7]),
]
MOVING_INDICES = BASIC_INDICES + OTHER_INDICES


@pytest.mark.parametrize("sources, dest, expectation", MOVING_INDICES)
def test_move_multiple(sources, dest, expectation):
    """Test the that we can move objects with the move method"""
    el = EventedList(range(8))
    el.events = Mock(wraps=el.events)
    assert el == [0, 1, 2, 3, 4, 5, 6, 7]

    def _fail():
        raise AssertionError("unexpected event called")

    el.events.removing.connect(_fail)
    el.events.removed.connect(_fail)
    el.events.inserting.connect(_fail)
    el.events.inserted.connect(_fail)

    el.move_multiple(sources, dest)
    assert el == expectation
    el.events.moving.emit.assert_called()
    el.events.moved.emit.assert_called()
    el.events.reordered.emit.assert_called()


def test_move_multiple_mimics_slice_reorder():
    """Test the that move_multiple provides the same result as slice insertion."""
    data = list(range(8))
    el = EventedList(data)
    el.events = Mock(wraps=el.events)
    assert el == data
    new_order = [1, 5, 3, 4, 6, 7, 2, 0]
    # this syntax
    el.move_multiple(new_order, 0)
    # is the same as this syntax
    data[:] = [data[i] for i in new_order]
    assert el == new_order
    assert el == data
    assert el.events.moving.emit.call_args_list == [
        call(1, 0),
        call(5, 1),
        call(4, 2),
        call(5, 3),
        call(6, 4),
        call(7, 5),
        call(7, 6),
    ]
    assert el.events.moved.emit.call_args_list == [
        call(1, 0, 1),
        call(5, 1, 5),
        call(4, 2, 3),
        call(5, 3, 4),
        call(6, 4, 6),
        call(7, 5, 7),
        call(7, 6, 2),
    ]
    el.events.reordered.emit.assert_called()

    # move_multiple also works omitting the insertion index
    el[:] = list(range(8))
    expected = [el[i] for i in new_order]
    el.move_multiple(new_order)
    assert el == expected


def test_child_events():
    """Test that evented lists bubble child events."""

    # create a random object that emits events
    class E:
        test = Signal(str)

    e_obj = E()
    root: EventedList[E] = EventedList(child_events=True)
    mock = Mock()
    root.events.all.connect(mock)
    root.append(e_obj)
    assert len(e_obj.test) == 1
    assert root == [e_obj]
    e_obj.test.emit("hi")

    assert mock.call_count == 3

    expected = [
        call(EmissionInfo(root.events.inserting, (0,), path=(PathStep(index=0),))),
        call(EmissionInfo(root.events.inserted, (0, e_obj), path=(PathStep(index=0),))),
        call(
            EmissionInfo(
                e_obj.test, ("hi",), path=(PathStep(index=0), PathStep(attr="test"))
            )
        ),
    ]
    mock.assert_has_calls(expected)

    del root[0]
    assert len(e_obj.test) == 0


def test_child_events_groups():
    """Test that evented lists bubble child events."""

    # create a random object that emits events
    class Group(SignalGroup):
        test = Signal(str)
        test2 = Signal(str)

    class E:
        def __init__(self):
            self.events = Group(self)

    e_obj = E()
    root: EventedList[E] = EventedList(child_events=True)
    mock = Mock()
    root.events.all.connect(mock)
    root.append(e_obj)
    assert root == [e_obj]
    e_obj.events.test2.emit("hi")

    assert [c[0][0].signal.name for c in mock.call_args_list] == [
        "inserting",
        "inserted",
        "test2",  # This is now the direct child signal, not child_event
    ]

    # when an object in the list owns an emitter group, then any emitter in that group
    # will also be detected, and the child event will be emitted directly with path info
    expected = [
        call(EmissionInfo(root.events.inserting, (0,), path=(PathStep(index=0),))),
        call(EmissionInfo(root.events.inserted, (0, e_obj), path=(PathStep(index=0),))),
        call(EmissionInfo(e_obj.events.test2, ("hi",), path=(PathStep(index=0),))),
    ]

    # note that we can get back to the actual object in the list using the .instance
    # attribute on signal instances.
    assert e_obj.events.test2.instance.all.instance == e_obj
    mock.assert_has_calls(expected)


def test_copy_no_sync():
    l1 = EventedList([1, 2, 3])
    l2 = copy(l1)
    l1.append(4)
    assert len(l2) == 3



import os
import sys
from pathlib import Path
from unittest.mock import Mock, call

import pytest

from psygnal import EmissionInfo, Signal, SignalGroup
from psygnal.utils import decompile, monitor_events, recompile


def test_event_debugger(capsys) -> None:
    """Test that the event debugger works"""

    class M:
        sig = Signal(int, int)

    m = M()
    _logger = Mock()

    assert not m.sig._slots

    with monitor_events(m, _logger):
        assert len(m.sig._slots) == 1
        m.sig.emit(1, 2)
        m.sig.emit(3, 4)

    assert _logger.call_count == 2
    _logger.assert_has_calls(
        [call(EmissionInfo(m.sig, (1, 2))), call(EmissionInfo(m.sig, (3, 4)))]
    )
    assert not m.sig._slots

    with monitor_events(m):
        m.sig.emit(1, 2)
        m.sig.emit(3, 4)

    captured = capsys.readouterr()
    assert captured.out == "sig.emit(1, 2)\nsig.emit(3, 4)\n"


def test_old_monitor_api_dep_warning() -> None:
    class M:
        sig = Signal(int, int)

    mock = Mock()

    def _monitor(signal_name: str, args: tuple) -> None:
        mock(signal_name, args)

    m = M()
    with pytest.warns(
        UserWarning, match="logger functions must now take a single argument"
    ):
        with monitor_events(m, logger=_monitor):  # type: ignore
            m.sig.emit(1, 2)
    mock.assert_called_once_with("sig", (1, 2))

    with pytest.raises(ValueError, match="logger function must take a single argument"):
        with monitor_events(logger=_monitor):  # type: ignore
            m.sig.emit(1, 2)

    mock.reset_mock()
    with monitor_events(m, logger=mock):
        m.sig.emit(1, 2)
    mock.assert_called_once_with(EmissionInfo(m.sig, (1, 2)))

    # global monitor
    mock.reset_mock()
    with monitor_events(logger=mock):
        m.sig.emit(1, 2)
    mock.assert_called_once_with(EmissionInfo(m.sig, (1, 2)))


def test_monitor_all() -> None:
    class M:
        sig = Signal(int, int)

    m1 = M()
    m2 = M()
    _logger = Mock()

    with monitor_events(logger=_logger):
        m1.sig.emit(1, 2)
        m2.sig.emit(3, 4)
        m1.sig.emit(5, 6)
        m2.sig.emit(7, 8)

    assert _logger.call_args_list == [
        call(EmissionInfo(m1.sig, (1, 2))),
        call(EmissionInfo(m2.sig, (3, 4))),
        call(EmissionInfo(m1.sig, (5, 6))),
        call(EmissionInfo(m2.sig, (7, 8))),
    ]


def test_monitor_group() -> None:
    class MyGroup(SignalGroup):
        sig1 = Signal(int, int)
        sig2 = Signal(str, str)

    m1 = MyGroup()
    m2 = MyGroup()
    _logger = Mock()

    with monitor_events(logger=_logger):
        m1.sig1.emit(1, 2)
        m2.sig1.emit(3, 4)
        m1.sig1.emit(5, 6)
        m2.sig1.emit(7, 8)
        m1.sig2.emit("9", "10")
        m2.sig2.emit("11", "12")

    assert _logger.call_args_list == [
        call(EmissionInfo(m1.sig1, (1, 2))),
        call(EmissionInfo(m2.sig1, (3, 4))),
        call(EmissionInfo(m1.sig1, (5, 6))),
        call(EmissionInfo(m2.sig1, (7, 8))),
        call(EmissionInfo(m1.sig2, ("9", "10"))),
        call(EmissionInfo(m2.sig2, ("11", "12"))),
    ]


@pytest.mark.skipif(os.name == "nt", reason="rewrite open files on Windows is buggy")
def test_decompile_recompile(monkeypatch):
    import psygnal

    was_compiled = psygnal._compiled

    decompile()
    monkeypatch.delitem(sys.modules, "psygnal")
    monkeypatch.delitem(sys.modules, "psygnal._signal")
    import psygnal

    assert not psygnal._compiled

    if was_compiled:
        assert list(Path(psygnal.__file__).parent.rglob("**/*_BAK"))
        recompile()
        monkeypatch.delitem(sys.modules, "psygnal")
        monkeypatch.delitem(sys.modules, "psygnal._signal")
        import psygnal

        assert psygnal._compiled


def test_debug_import(monkeypatch):
    """Test that PSYGNAL_UNCOMPILED gives a warning."""
    monkeypatch.delitem(sys.modules, "psygnal")
    monkeypatch.setenv("PSYGNAL_UNCOMPILED", "1")

    with pytest.warns(UserWarning, match="PSYGNAL_UNCOMPILED no longer has any effect"):
        import psygnal  # noqa: F401



from typing import Any, get_origin

import pytest

try:
    import pydantic
except ImportError:
    pytest.skip("pydantic not installed", allow_module_level=True)

from psygnal import containers

V1 = pydantic.__version__.startswith("1")


@pytest.mark.skipif(V1, reason="pydantic v1 has poor support for generics")
@pytest.mark.parametrize(
    "hint",
    [
        containers.EventedList[int],
        containers.SelectableEventedList[int],
    ],
)
def test_evented_list_as_pydantic_field(hint: Any) -> None:
    class Model(pydantic.BaseModel):
        my_list: hint

    m = Model(my_list=[1, 2, 3])  # type: ignore
    assert m.my_list == [1, 2, 3]
    assert isinstance(m.my_list, get_origin(hint))

    m2 = Model(my_list=containers.EventedList([1, 2, 3]))
    assert m2.my_list == [1, 2, 3]
    m3 = Model(my_list=[1, "2", 3])  # type: ignore
    assert m3.my_list == [1, 2, 3]
    assert isinstance(m3.my_list, get_origin(hint))

    with pytest.raises(pydantic.ValidationError):
        Model(my_list=[1, 2, "string"])  # type: ignore


@pytest.mark.skipif(V1, reason="pydantic v1 has poor support for generics")
def test_evented_list_no_params_as_pydantic_field() -> None:
    class Model(pydantic.BaseModel):
        my_list: containers.EventedList

    m = Model(my_list=[1, 2, 3])  # type: ignore
    assert m.my_list == [1, 2, 3]
    assert isinstance(m.my_list, containers.EventedList)

    m3 = Model(my_list=[1, "string", 3])  # type: ignore
    assert m3.my_list == [1, "string", 3]
    assert isinstance(m3.my_list, containers.EventedList)


@pytest.mark.skipif(V1, reason="pydantic v1 has poor support for generics")
@pytest.mark.parametrize(
    "hint",
    [
        containers.EventedSet[str],
        containers.EventedOrderedSet[str],
        containers.Selection[str],
    ],
)
def test_evented_set_as_pydantic_field(hint: Any) -> None:
    class Model(pydantic.BaseModel):
        my_set: hint

        model_config = {"coerce_numbers_to_str": True}

    m = Model(my_set=[1, 2])  # type: ignore
    assert m.my_set == {"1", "2"}  # type: ignore
    assert isinstance(m.my_set, get_origin(hint))

    m2 = Model(my_set=containers.EventedSet(["a", "b"]))
    assert m2.my_set == {"a", "b"}  # type: ignore
    m3 = Model(my_set=[1, "2", 3])  # type: ignore
    assert m3.my_set == {"1", "2", "3"}  # type: ignore
    assert isinstance(m3.my_set, get_origin(hint))


@pytest.mark.skipif(V1, reason="pydantic v1 has poor support for generics")
def test_evented_dict_as_pydantic_field() -> None:
    class Model(pydantic.BaseModel):
        my_dict: containers.EventedDict[str, int]

        model_config = {"coerce_numbers_to_str": True}

    m = Model(my_dict={"a": 1})  # type: ignore
    assert m.my_dict == {"a": 1}
    assert isinstance(m.my_dict, containers.EventedDict)

    m2 = Model(my_dict=containers.EventedDict({"a": 1}))
    assert m2.my_dict == {"a": 1}
    assert isinstance(m2.my_dict, containers.EventedDict)

    m3 = Model(my_dict={1: "2"})  # type: ignore
    assert m3.my_dict == {"1": 2}
    assert isinstance(m3.my_dict, containers.EventedDict)

    with pytest.raises(pydantic.ValidationError):
        Model(my_dict={"a": "string"})  # type: ignore



from copy import copy
from unittest.mock import Mock, call

import pytest

from psygnal.containers import EventedOrderedSet, EventedSet, OrderedSet


@pytest.fixture
def regular_set():
    return set(range(5))


@pytest.fixture(params=[EventedSet, EventedOrderedSet])
def test_set(request, regular_set):
    return request.param(regular_set)


@pytest.mark.parametrize(
    "meth",
    [
        # METHOD, ARGS, EXPECTED EVENTS
        # primary interface
        ("add", 2, []),
        ("add", 10, [call((10,), ())]),
        ("discard", 2, [call((), (2,))]),
        ("remove", 2, [call((), (2,))]),
        ("discard", 10, []),
        # parity with set
        ("update", {3, 4, 5, 6}, [call((5, 6), ())]),
        ("difference_update", {3, 4, 5, 6}, [call((), (3, 4))]),
        ("intersection_update", {3, 4, 5, 6}, [call((), (0, 1, 2))]),
        ("symmetric_difference_update", {3, 4, 5, 6}, [call((5, 6), (3, 4))]),
    ],
    ids=lambda x: x[0],
)
def test_set_interface_parity(test_set: EventedSet, regular_set: set, meth):
    method_name, arg, expected = meth
    mock = Mock()

    test_set.events.items_changed.connect(mock)

    test_set_method = getattr(test_set, method_name)
    assert tuple(test_set) == tuple(regular_set)

    regular_set_method = getattr(regular_set, method_name)
    assert test_set_method(arg) == regular_set_method(arg)
    assert tuple(test_set) == tuple(regular_set)

    mock.assert_has_calls(expected)
    assert type(test_set).__name__ in repr(test_set)


def test_set_pop(test_set: EventedSet):
    mock = Mock()
    test_set.events.items_changed.connect(mock)

    npops = len(test_set)
    while test_set:
        test_set.pop()

    assert mock.call_count == npops

    with pytest.raises(KeyError):
        test_set.pop()
    with pytest.raises(KeyError):
        test_set.remove(34)


def test_set_clear(test_set: EventedSet):
    mock = Mock()
    test_set.events.items_changed.connect(mock)
    mock.assert_not_called()
    test_set.clear()
    mock.assert_called_once_with((), (0, 1, 2, 3, 4))


@pytest.mark.parametrize(
    "meth",
    [
        ("difference", {3, 4, 5, 6}),
        ("intersection", {3, 4, 5, 6}),
        ("issubset", {3, 4}),
        ("issubset", {3, 4, 5, 6}),
        ("issubset", {1, 2, 3, 4, 5, 6}),
        ("issuperset", {3, 4}),
        ("issuperset", {3, 4, 5, 6}),
        ("issuperset", {1, 2, 3, 4, 5, 6}),
        ("symmetric_difference", {3, 4, 5, 6}),
        ("union", {3, 4, 5, 6}),
    ],
)
def test_set_new_objects(test_set: EventedSet, regular_set: set, meth):
    method_name, arg = meth
    test_set_method = getattr(test_set, method_name)
    assert tuple(test_set) == tuple(regular_set)

    mock = Mock()
    test_set.events.items_changed.connect(mock)
    regular_set_method = getattr(regular_set, method_name)
    result = test_set_method(arg)
    assert result == regular_set_method(arg)
    assert isinstance(result, (EventedSet, EventedOrderedSet, bool))
    assert result is not test_set

    mock.assert_not_called()


def test_ordering():
    tup = (24, 16, 8, 4, 5, 6)
    s_tup = set(tup)
    os_tup = OrderedSet(tup)

    assert tuple(s_tup) != tup
    assert repr(s_tup) == "{4, 5, 6, 8, 16, 24}"

    assert tuple(os_tup) == tup
    assert repr(os_tup) == "OrderedSet((24, 16, 8, 4, 5, 6))"
    os_tup.discard(8)
    os_tup.add(8)
    assert tuple(os_tup) == (24, 16, 4, 5, 6, 8)


def test_copy(test_set):
    from copy import copy

    assert test_set.copy() == copy(test_set)
    assert test_set is not copy(test_set)
    assert isinstance(copy(test_set), type(test_set))


def test_repr(test_set):
    if isinstance(test_set, EventedOrderedSet):
        assert repr(test_set) == "EventedOrderedSet((0, 1, 2, 3, 4))"
    else:
        assert repr(test_set) == "EventedSet({0, 1, 2, 3, 4})"


def test_copy_no_sync():
    s1 = EventedSet([1, 2, 3])
    s2 = copy(s1)
    s1.add(4)
    assert len(s2) == 3


def test_set_emission_order():
    s = EventedSet()

    def callback1():
        if 1 not in s:
            s.add(1)

    def callback2():
        if 5 not in s:
            s.update(range(5, 10))

    s.events.items_changed.connect(callback1)
    s.events.items_changed.connect(callback2)
    mock = Mock()
    s.events.items_changed.connect(mock)

    s.add(11)
    mock.assert_has_calls(
        [
            call((11,), ()),
            call((1,), ()),
            call((5, 6, 7, 8, 9), ()),
        ]
    )



"""Tests to cover missing lines in path/emission functionality."""

from inspect import Signature

import pytest

from psygnal import EmissionInfo, PathStep, Signal, SignalGroup, SignalInstance


def test_pathstep_validation():
    """Test PathStep validation error conditions."""
    # Test empty PathStep (should fail)
    with pytest.raises(ValueError, match="exactly one of attr, index, or key"):
        PathStep()

    # Test multiple fields set (should fail)
    with pytest.raises(ValueError, match="exactly one of attr, index, or key"):
        PathStep(attr="test", index=1)


def test_pathstep_repr():
    """Test PathStep repr formatting, including long key truncation."""
    # Test attr
    assert repr(PathStep(attr="test")) == ".test"

    # Test index
    assert repr(PathStep(index=5)) == "[5]"

    # Test short key
    assert repr(PathStep(key="short")) == "['short']"

    # Test long key truncation
    class CrazyHashable:
        """A class with a long __repr__."""

        def __repr__(self):
            return "a" * 100

    ps = PathStep(key=CrazyHashable())
    result = repr(ps)
    assert "..." in result
    assert len(result) <= 25  # should be truncated


def test_emission_info_path_validation():
    """Test EmissionInfo path validation."""
    # Create a SignalInstance properly
    instance = SignalInstance(Signature())

    # Valid paths should work
    EmissionInfo(instance, (1,), (PathStep(attr="test"),))
    EmissionInfo(instance, (1,), (PathStep(index=0), PathStep(key="key")))

    # Invalid path types should fail
    with pytest.raises(TypeError):
        EmissionInfo(instance, (1,), (object(),))  # type: ignore


def test_signal_relay_no_emitter():
    """Test SignalRelay when no current emitter."""

    class TestGroup(SignalGroup):
        test_signal = Signal(int)

    group = TestGroup()

    # Test that _slot_relay returns early when no current emitter
    # This should not raise an error and should not emit anything
    relay = group.all
    relay._slot_relay(1, 2, 3)  # Should return early due to no emitter


def test_signal_group_repr_without_instance():
    """Test SignalGroup repr when instance is None."""

    class TestGroup(SignalGroup):
        test_signal = Signal(int)

    # Create group without instance
    group = TestGroup()
    repr_str = repr(group)
    assert "TestGroup" in repr_str
    assert "instance at" not in repr_str  # Should not have instance info


def test_signal_group_repr_with_instance():
    """Test SignalGroup repr when instance is not None."""

    class TestGroup(SignalGroup):
        test_signal = Signal(int)

    class TestObj:
        def __init__(self):
            self.events = TestGroup(instance=self)

    # Create group with instance
    obj = TestObj()
    repr_str = repr(obj.events)
    assert "TestGroup" in repr_str
    assert "instance at" in repr_str  # Should have instance info
    assert "TestObj" in repr_str


def test_list_signal_instance_relocate_empty_args():
    """Test ListSignalInstance _psygnal_relocate_info_ with empty args."""
    from psygnal.containers._evented_list import ListSignalInstance

    # Create a signal instance
    list_sig = ListSignalInstance((int,))

    # Test with empty args
    info = EmissionInfo(list_sig, ())
    result = list_sig._psygnal_relocate_info_(info)
    assert result is info  # Should return unchanged


def test_dict_signal_instance_relocate_empty_args():
    """Test DictSignalInstance _psygnal_relocate_info_ with empty args."""
    from psygnal.containers._evented_dict import DictSignalInstance

    # Create a signal instance
    dict_sig = DictSignalInstance((str,))

    # Test with empty args
    info = EmissionInfo(dict_sig, ())
    result = dict_sig._psygnal_relocate_info_(info)
    assert result is info  # Should return unchanged



from copy import copy
from unittest.mock import Mock

import pytest

from psygnal.containers._evented_dict import EventedDict


@pytest.fixture
def regular_dict():
    return {"A": 1, "B": 2, "C": 3}


@pytest.fixture
def test_dict(regular_dict):
    """EventedDict without basetype set."""
    test_dict = EventedDict(regular_dict)
    test_dict.events = Mock(wraps=test_dict.events)
    return test_dict


@pytest.mark.parametrize(
    "method_name, args, expected",
    [
        ("__getitem__", ("A",), 1),  # read
        ("__setitem__", ("A", 3), None),  # update
        ("__setitem__", ("D", 3), None),  # add new entry
        ("__delitem__", ("A",), None),  # delete
        ("__len__", (), 3),
        ("__newlike__", ({"A": 1},), {"A": 1}),
        ("copy", (), {"A": 1, "B": 2, "C": 3}),
    ],
)
def test_dict_interface_parity(regular_dict, test_dict, method_name, args, expected):
    """Test that EventedDict interface is equivalent to the builtin dict."""
    test_dict_method = getattr(test_dict, method_name)
    assert test_dict == regular_dict
    if hasattr(regular_dict, method_name):
        regular_dict_method = getattr(regular_dict, method_name)
        assert test_dict_method(*args) == regular_dict_method(*args) == expected
        assert test_dict == regular_dict
    else:
        test_dict_method(*args)  # smoke test


def test_dict_inits():
    a = EventedDict({"A": 1, "B": 2, "C": 3})
    b = EventedDict(A=1, B=2, C=3)
    c = EventedDict({"A": 1}, B=2, C=3)
    assert a == b == c


def test_dict_repr(test_dict):
    assert repr(test_dict) == "EventedDict({'A': 1, 'B': 2, 'C': 3})"


def test_instantiation_without_data():
    """Test that EventedDict can be instantiated without data."""
    test_dict = EventedDict()
    assert isinstance(test_dict, EventedDict)


def test_basetype_enforcement_on_instantiation():
    """EventedDict with basetype set should enforce types on instantiation."""
    with pytest.raises(TypeError):
        test_dict = EventedDict({"A": "not an int"}, basetype=int)
    test_dict = EventedDict({"A": 1})
    assert isinstance(test_dict, EventedDict)


def test_basetype_enforcement_on_set_item():
    """EventedDict with basetype set should enforces types on setitem."""
    test_dict = EventedDict(basetype=int)
    test_dict["A"] = 1
    with pytest.raises(TypeError):
        test_dict["A"] = "not an int"


def test_dict_add_events(test_dict):
    """Test that events are emitted before and after an item is added."""
    test_dict.events.adding.emit = Mock(wraps=test_dict.events.adding.emit)
    test_dict.events.added.emit = Mock(wraps=test_dict.events.added.emit)
    test_dict["D"] = 4
    test_dict.events.adding.emit.assert_called_with("D")
    test_dict.events.added.emit.assert_called_with("D", 4)

    test_dict.events.adding.emit.reset_mock()
    test_dict.events.added.emit.reset_mock()
    test_dict["D"] = 4
    test_dict.events.adding.emit.assert_not_called()
    test_dict.events.added.emit.assert_not_called()


def test_dict_change_events(test_dict):
    """Test that events are emitted when an item in the dictionary is replaced."""
    # events shouldn't be emitted on addition

    test_dict.events.changing.emit = Mock(wraps=test_dict.events.changing.emit)
    test_dict.events.changed.emit = Mock(wraps=test_dict.events.changed.emit)
    test_dict["D"] = 4
    test_dict.events.changing.emit.assert_not_called()
    test_dict.events.changed.emit.assert_not_called()
    test_dict["C"] = 4
    test_dict.events.changing.emit.assert_called_with("C")
    test_dict.events.changed.emit.assert_called_with("C", 3, 4)


def test_dict_remove_events(test_dict):
    """Test that events are emitted before and after an item is removed."""
    test_dict.events.removing.emit = Mock(wraps=test_dict.events.removing.emit)
    test_dict.events.removed.emit = Mock(wraps=test_dict.events.removed.emit)
    test_dict.pop("C")
    test_dict.events.removing.emit.assert_called_with("C")
    test_dict.events.removed.emit.assert_called_with("C", 3)


def test_copy_no_sync():
    d1 = EventedDict({1: 1, 2: 2, 3: 3})
    d2 = copy(d1)
    d1[4] = 4
    d1[3] = 4
    assert len(d2) == 3
    assert d2[3] == 3



from unittest.mock import Mock

import pytest

from psygnal.containers import SelectableEventedList


@pytest.fixture
def regular_list() -> list:
    return list(range(5))


@pytest.fixture
def test_list(regular_list: list) -> SelectableEventedList:
    test_list = SelectableEventedList(regular_list)
    test_list.events = Mock(wraps=test_list.events)
    test_list.selection.events = Mock(wraps=test_list.selection.events)
    return test_list


def test_select_item_not_in_list(test_list: SelectableEventedList) -> None:
    """Items not in list should not be added to selection."""
    with pytest.raises(ValueError):
        test_list.selection.add(6)
    assert 6 not in test_list.selection


def test_newly_selected_item_is_active(test_list: SelectableEventedList) -> None:
    """Items added to a selection should become active."""
    test_list.selection.clear()
    test_list.selection.add(1)
    assert test_list.selection.active == 1


def test_select_all(test_list: SelectableEventedList) -> None:
    """Select all should populate the selection."""
    test_list.selection.update = Mock(wraps=test_list.selection.update)
    test_list.selection.clear()
    assert not test_list.selection
    test_list.select_all()
    assert all(el in test_list.selection for el in range(5))
    test_list.selection.update.assert_called_once()


def test_deselect_all(test_list: SelectableEventedList) -> None:
    """Deselect all should clear the selection"""
    test_list.selection.clear = Mock(wraps=test_list.selection.clear)
    test_list.selection = list(range(5))
    assert all(el in test_list.selection for el in range(5))
    test_list.deselect_all()
    assert not test_list.selection
    test_list.selection.clear.assert_called_once()


@pytest.mark.parametrize(
    "initial_selection, step, expand_selection, wraparound, expected",
    [
        ({0}, 1, False, False, {1}),
        ({0}, 2, False, False, {2}),
        ({0}, 2, True, False, {0, 2}),
        ({0, 1}, 1, False, False, {1}),
        ({0}, 5, False, False, {4}),
        ({0}, 5, False, True, {0}),
        ({}, 1, False, False, {4}),
    ],
)
def test_select_next(
    test_list: SelectableEventedList,
    initial_selection,
    step,
    expand_selection,
    wraparound,
    expected,
):
    """Test select next method behaviour."""
    test_list.selection = initial_selection
    test_list.select_next(
        step=step, expand_selection=expand_selection, wraparound=wraparound
    )
    assert test_list.selection == expected


def test_select_next_with_empty_list():
    """Selection should remain unchanged on advancing if list is empty."""
    test_list = SelectableEventedList([])
    initial_selection = test_list.selection.copy()
    test_list.select_next()
    assert test_list.selection == initial_selection


@pytest.mark.parametrize(
    "initial_selection, expand_selection, wraparound, expected",
    [
        ({1}, False, False, {0}),
        ({0}, False, False, {0}),
        ({1}, True, False, {0, 1}),
        ({1, 2}, False, False, {0}),
        ({0}, False, True, {4}),
    ],
)
def test_select_previous(
    test_list, initial_selection, expand_selection, wraparound, expected
):
    """Test select next method behaviour."""
    test_list.selection = initial_selection
    test_list.select_previous(expand_selection=expand_selection, wraparound=wraparound)
    assert test_list.selection == expected


def test_item_discarded_from_selection_on_removal_from_list(
    test_list: SelectableEventedList,
) -> None:
    """Check that items removed from a list are also removed from the selection."""
    test_list.selection.clear()
    test_list.selection.discard = Mock(wraps=test_list.selection.discard)
    test_list.selection = {0}
    assert 0 in test_list.selection
    test_list.remove(0)
    assert 0 not in test_list.selection
    test_list.selection.discard.assert_called_once()


def test_remove_selected(test_list: SelectableEventedList) -> None:
    """Test items are removed from both the selection and the list."""
    test_list.selection.clear()
    initial_selection = {0, 1}
    test_list.selection = initial_selection
    assert test_list.selection == initial_selection

    output = test_list.remove_selected()
    assert set(output) == initial_selection
    assert all(el not in test_list for el in initial_selection)
    assert all(el not in test_list.selection for el in initial_selection)
    assert test_list.selection == {2}



import gc
import os
from contextlib import suppress
from functools import partial, wraps
from inspect import Signature
from typing import Literal
from unittest.mock import MagicMock, Mock, call

import pytest

import psygnal
from psygnal import EmitLoopError, Signal, SignalInstance
from psygnal._signal import ReemissionMode, ReemissionVal
from psygnal._weak_callback import WeakCallback

WINDOWS = os.name == "nt"
COMPILED = psygnal._compiled


def stupid_decorator(fun):
    def _fun(*args):
        fun(*args)

    _fun.__annotations__ = fun.__annotations__
    _fun.__name__ = "f_no_arg"
    return _fun


def good_decorator(fun):
    @wraps(fun)
    def _fun(*args):
        fun(*args)

    return _fun


# fmt: off
class Emitter:
    no_arg = Signal()
    one_int = Signal(int)
    two_int = Signal(int, int)
    str_int = Signal(str, int)
    no_check = Signal(str, check_nargs_on_connect=False, check_types_on_connect=False)


class MyObj:
    def f_no_arg(self): ...
    def f_str_int_vararg(self, a: str, b: int, *c): ...
    def f_str_int_any(self, a: str, b: int, c): ...
    def f_str_int_kwarg(self, a: str, b: int, c=None): ...
    def f_str_int(self, a: str, b: int): ...
    def f_str_any(self, a: str, b): ...
    def f_str(self, a: str): ...
    def f_int(self, a: int): ...
    def f_any(self, a): ...
    def f_int_int(self, a: int, b: int): ...
    def f_str_str(self, a: str, b: str): ...
    def f_arg_kwarg(self, a, b=None): ...
    def f_vararg(self, *a): ...
    def f_vararg_varkwarg(self, *a, **b): ...
    def f_vararg_kwarg(self, *a, b=None): ...
    @stupid_decorator
    def f_int_decorated_stupid(self, a: int): ...
    @good_decorator
    def f_int_decorated_good(self, a: int): ...
    f_any_assigned = lambda self, a: None  # noqa

    x: int = 0
    def __setitem__(self, key: str, value: int):
        if key == "x":
            self.x = value

def f_no_arg(): ...
def f_str_int_vararg(a: str, b: int, *c): ...
def f_str_int_any(a: str, b: int, c): ...
def f_str_int_kwarg(a: str, b: int, c=None): ...
def f_str_int(a: str, b: int): ...
def f_str_any(a: str, b): ...
def f_str(a: str): ...
def f_int(a: int): ...
def f_any(a): ...
def f_int_int(a: int, b: int): ...
def f_str_str(a: str, b: str): ...
def f_arg_kwarg(a, b=None): ...
def f_vararg(*a): ...
def f_vararg_varkwarg(*a, **b): ...
def f_vararg_kwarg(*a, b=None): ...

class MyReceiver:
    expect_signal = None
    expect_sender = None
    expect_name = None

    def assert_sender(self, *a):
        assert Signal.current_emitter() is self.expect_signal
        assert self.expect_name in repr(Signal.current_emitter())
        assert Signal.current_emitter().instance is self.expect_sender
        assert Signal.sender() is self.expect_sender
        assert Signal.current_emitter()._name is self.expect_name

    def assert_not_sender(self, *a):
        # just to make sure we're actually calling it
        assert Signal.current_emitter().instance is not self.expect_sender
# fmt: on


def test_basic_signal():
    """standard Qt usage, as class attribute"""
    emitter = Emitter()
    mock = MagicMock()
    emitter.one_int.connect(mock)
    emitter.one_int.emit(1)
    mock.assert_called_once_with(1)
    mock.reset_mock()

    # calling directly also works
    emitter.one_int(1)
    mock.assert_called_once_with(1)


def test_emit_fast():
    """Test emit_fast method."""
    emitter = Emitter()
    mock = MagicMock()
    emitter.one_int.connect(mock)
    emitter.one_int.emit_fast(1)
    mock.assert_called_once_with(1)
    mock.reset_mock()

    # calling directly also works
    emitter.one_int(1)
    mock.assert_called_once_with(1)
    mock.reset_mock()

    with emitter.one_int.blocked():
        emitter.one_int.emit_fast(2)
    mock.assert_not_called()

    with emitter.one_int.paused():
        emitter.one_int.emit_fast(3)
        mock.assert_not_called()
        emitter.one_int.emit_fast(4)
    mock.assert_has_calls([call(3), call(4)])


def test_emit_fast_errors():
    emitter = Emitter()
    err = ValueError()

    @emitter.one_int.connect
    def boom(v: int) -> None:
        raise err

    import re

    error_re = re.compile(
        f"signal 'tests.test_psygnal.Emitter.one_int'.*{re.escape(__file__)}",
        re.DOTALL,
    )
    with pytest.raises(EmitLoopError, match=error_re):
        emitter.one_int.emit_fast(42)


def test_emit_fast_recursion_errors():
    """Test emit_fast method."""
    emitter = Emitter()
    emitter.one_int.emit_fast(1)

    @emitter.one_int.connect
    def callback() -> None:
        emitter.one_int.emit(2)

    with pytest.raises(RecursionError):
        emitter.one_int.emit_fast(3)

    emitter.one_int.disconnect(callback)

    @emitter.one_int.connect
    def callback() -> None:
        emitter.one_int.emit_fast(2)

    with pytest.raises(RecursionError):
        emitter.one_int.emit_fast(3)


def test_decorator():
    emitter = Emitter()
    err = ValueError()

    @emitter.one_int.connect
    def boom(v: int) -> None:
        raise err

    @emitter.one_int.connect(check_nargs=False)
    def bad_cb(a, b, c): ...

    import re

    error_re = re.compile(
        f"signal 'tests.test_psygnal.Emitter.one_int'.*{re.escape(__file__)}",
        re.DOTALL,
    )
    with pytest.raises(EmitLoopError, match=error_re) as e:
        emitter.one_int.emit(42)

    assert e.value.__cause__ is err
    assert e.value.__context__ is err


def test_misc():
    emitter = Emitter()
    assert isinstance(Emitter.one_int, Signal)
    assert isinstance(emitter.one_int, SignalInstance)

    with pytest.raises(AttributeError):
        _ = emitter.one_int.asdf

    with pytest.raises(AttributeError):
        _ = emitter.one_int.asdf


def test_getattr():
    s = Signal()
    with pytest.raises(AttributeError):
        _ = s.not_a_thing


def test_signature_provided():
    s = Signal(Signature())
    assert s.signature == Signature()

    with pytest.warns(UserWarning):
        s = Signal(Signature(), 1)


def test_emit_checks():
    emitter = Emitter()
    emitter.one_int.connect(f_no_arg)

    emitter.one_int.emit(check_nargs=False)
    emitter.one_int.emit()
    with pytest.raises(TypeError):
        emitter.one_int.emit(check_nargs=True)

    emitter.one_int.emit(1)

    emitter.one_int.emit(1, 2, check_nargs=False)
    emitter.one_int.emit(1, 2)
    with pytest.raises(TypeError):
        emitter.one_int.emit(1, 2, check_nargs=True)

    with pytest.raises(TypeError):
        emitter.one_int.emit("sdr", check_types=True)

    emitter.one_int.emit("sdr", check_types=False)


def test_basic_signal_blocked():
    """standard Qt usage, as class attribute"""
    emitter = Emitter()
    mock = MagicMock()

    emitter.one_int.connect(mock)
    emitter.one_int.emit(1)
    mock.assert_called_once_with(1)

    mock.reset_mock()
    with emitter.one_int.blocked():
        emitter.one_int.emit(1)
    mock.assert_not_called()


def test_nested_signal_blocked():
    """unblock signal on exit of the last context"""
    emitter = Emitter()
    mock = MagicMock()

    emitter.one_int.connect(mock)

    mock.reset_mock()
    with emitter.one_int.blocked():
        with emitter.one_int.blocked():
            emitter.one_int.emit(1)
        emitter.one_int.emit(2)
    emitter.one_int.emit(3)
    mock.assert_called_once_with(3)


@pytest.mark.parametrize("thread", [None, "main"])
def test_disconnect(thread: Literal[None, "main"]) -> None:
    emitter = Emitter()
    mock = MagicMock()
    with pytest.raises(ValueError) as e:
        emitter.one_int.disconnect(mock, missing_ok=False)
    assert "slot is not connected" in str(e)
    emitter.one_int.disconnect(mock)

    emitter.one_int.connect(mock, thread=thread)
    assert len(emitter.one_int) == 1
    if thread is None:
        emitter.one_int.emit(1)
        mock.assert_called_once_with(1)
        mock.reset_mock()

    emitter.one_int.disconnect(mock)
    emitter.one_int.emit(1)
    mock.assert_not_called()
    assert len(emitter.one_int) == 0


@pytest.mark.parametrize(
    "type_",
    [
        "function",
        "lambda",
        "method",
        "partial_method",
        "toolz_function",
        "toolz_method",
        "partial_method_kwarg",
        "partial_method_kwarg_bad",
        "setattr",
        "setitem",
    ],
)
def test_slot_types(type_: str) -> None:
    emitter = Emitter()
    signal = emitter.one_int
    assert len(signal) == 0
    obj = MyObj()

    if type_ == "setattr":
        with pytest.warns(FutureWarning, match="The default value of maxargs"):
            signal.connect_setattr(obj, "x")
    elif type_ == "setitem":
        with pytest.warns(FutureWarning, match="The default value of maxargs"):
            signal.connect_setitem(obj, "x")
    elif type_ == "function":
        signal.connect(f_int)

    elif type_ == "lambda":
        signal.connect(lambda x: None)
    elif type_ == "method":
        signal.connect(obj.f_int)
    elif type_ == "partial_method":
        signal.connect(partial(obj.f_int_int, 2))
    elif type_ == "toolz_function":
        toolz = pytest.importorskip("toolz")
        signal.connect(toolz.curry(f_int_int, 2))
    elif type_ == "toolz_method":
        toolz = pytest.importorskip("toolz")
        signal.connect(toolz.curry(obj.f_int_int, 2))
    elif type_ == "partial_method_kwarg":
        signal.connect(partial(obj.f_int_int, b=2))
    elif type_ == "partial_method_kwarg_bad":
        with pytest.raises(ValueError, match=r".*prefer using positional args"):
            signal.connect(partial(obj.f_int_int, a=2))
        return

    assert len(signal) == 1

    stored_slot = signal._slots[-1]
    assert isinstance(stored_slot, WeakCallback)
    assert stored_slot == stored_slot

    with pytest.raises(TypeError):
        emitter.one_int.connect("not a callable")  # type: ignore


def test_basic_signal_with_sender_receiver():
    """standard Qt usage, as class attribute"""
    emitter = Emitter()
    receiver = MyReceiver()
    receiver.expect_sender = emitter
    receiver.expect_signal = emitter.one_int
    receiver.expect_name = "one_int"

    assert Signal.current_emitter() is None
    emitter.one_int.connect(receiver.assert_sender)
    emitter.one_int.emit(1)

    # back to none after the call is over.
    assert Signal.current_emitter() is None
    emitter.one_int.disconnect()

    # sanity check... to make sure that methods are in fact being called.
    emitter.one_int.connect(receiver.assert_not_sender)
    with pytest.raises(EmitLoopError) as e:
        emitter.one_int.emit(1)

    assert isinstance(e.value.__cause__, AssertionError)
    assert isinstance(e.value.__context__, AssertionError)


def test_basic_signal_with_sender_nonreceiver():
    """standard Qt usage, as class attribute"""

    emitter = Emitter()
    nr = MyObj()

    emitter.one_int.connect(nr.f_no_arg)
    emitter.one_int.connect(nr.f_int)
    emitter.one_int.connect(nr.f_vararg_varkwarg)
    emitter.one_int.emit(1)

    # emitter.one_int.connect(nr.two_int)


def test_signal_instance():
    """make a signal instance without a class"""
    signal = SignalInstance((int,))
    mock = MagicMock()
    signal.connect(mock)
    signal.emit(1)
    mock.assert_called_once_with(1)

    signal = SignalInstance()
    mock = MagicMock()
    signal.connect(mock)
    signal.emit()
    mock.assert_called_once_with()


@pytest.mark.parametrize(
    "slot",
    [
        "f_no_arg",
        "f_int_decorated_stupid",
        "f_int_decorated_good",
        "f_any_assigned",
        "partial",
        "toolz_curry",
    ],
)
def test_weakref(slot):
    """Test that a connected method doesn't hold strong ref."""
    emitter = Emitter()
    obj = MyObj()

    assert len(emitter.one_int) == 0
    if slot == "partial":
        emitter.one_int.connect(partial(obj.f_int_int, 1))
    elif slot == "toolz_curry":
        toolz = pytest.importorskip("toolz")
        emitter.one_int.connect(toolz.curry(obj.f_int_int, 1))
    else:
        emitter.one_int.connect(getattr(obj, slot))
    assert len(emitter.one_int) == 1
    emitter.one_int.emit(1)
    assert len(emitter.one_int) == 1
    del obj
    gc.collect()
    emitter.one_int.emit(1)  # this should trigger deletion
    assert len(emitter.one_int) == 0


@pytest.mark.parametrize(
    "slot",
    [
        "f_no_arg",
        "f_int_decorated_stupid",
        "f_int_decorated_good",
        "f_any_assigned",
        "partial",
    ],
)
def test_group_weakref(slot: str) -> None:
    """Test that a connected method doesn't hold strong ref."""
    from psygnal import SignalGroup

    class MyGroup(SignalGroup):
        sig1 = Signal(int)

    group = MyGroup()
    obj = MyObj()

    assert len(group.sig1) == 0
    # but the group itself doesn't have any
    assert len(group._psygnal_relay) == 0

    # connecting something to the group adds to the group connections
    group.all.connect(
        partial(obj.f_int_int, 1) if slot == "partial" else getattr(obj, slot)
    )
    assert len(group.sig1) == 1
    assert len(group._psygnal_relay) == 1

    group.sig1.emit(1)
    assert len(group.sig1) == 1
    del obj
    gc.collect()
    group.sig1.emit(1)  # this should trigger deletion, so would emitter.emit()
    assert len(group.sig1) == 0  # NOTE! this is 0 not 1, because the relay is also gone
    assert len(group._psygnal_relay) == 0  # it's been cleaned up


# def test_norm_slot():
#     r = MyObj()
#     normed1 = _normalize_slot(r.f_any)
#     normed2 = _normalize_slot(normed1)
#     normed3 = _normalize_slot((r, "f_any", None))
#     normed4 = _normalize_slot((weakref.ref(r), "f_any", None))
#     assert normed1 == (weakref.ref(r), "f_any", None)
#     assert normed1 == normed2 == normed3 == normed4
#     assert _normalize_slot(f_any) == f_any


ALL = {n for n, f in locals().items() if callable(f) and n.startswith("f_")}
COUNT_INCOMPATIBLE = {
    "no_arg": ALL - {"f_no_arg", "f_vararg", "f_vararg_varkwarg", "f_vararg_kwarg"},
    "one_int": {
        "f_int_int",
        "f_str_any",
        "f_str_int_any",
        "f_str_int_kwarg",
        "f_str_int_vararg",
        "f_str_int",
        "f_str_str",
    },
    "str_int": {"f_str_int_any"},
}

SIG_INCOMPATIBLE = {
    "no_arg": {"f_int_int", "f_int", "f_str_int_any", "f_str_str"},
    "one_int": {
        "f_int_int",
        "f_str_int_any",
        "f_str_int_vararg",
        "f_str_str",
        "f_str",
    },
    "str_int": {"f_int_int", "f_int", "f_str_int_any", "f_str_str"},
}


@pytest.mark.parametrize("typed", ["typed", "untyped"])
@pytest.mark.parametrize("func_name", ALL)
@pytest.mark.parametrize("sig_name", ["no_arg", "one_int", "str_int"])
@pytest.mark.parametrize("mode", ["func", "meth", "partial"])
def test_connect_validation(func_name, sig_name, mode, typed):
    from functools import partial

    if mode == "meth":
        func = getattr(MyObj(), func_name)
    elif mode == "partial":
        func = partial(globals()[func_name])
    else:
        func = globals()[func_name]
    e = Emitter()

    check_types = typed == "typed"
    signal: SignalInstance = getattr(e, sig_name)
    bad_count = COUNT_INCOMPATIBLE[sig_name]
    bad_sig = SIG_INCOMPATIBLE[sig_name]
    if func_name in bad_count or (check_types and func_name in bad_sig):
        with pytest.raises(ValueError) as er:
            signal.connect(func, check_types=check_types)
        assert "Accepted signature:" in str(er)
        return

    signal.connect(func, check_types=check_types)

    args = (p.annotation() for p in signal.signature.parameters.values())
    signal.emit(*args)


def test_connect_lambdas():
    e = Emitter()
    assert len(e.two_int._slots) == 0
    e.two_int.connect(lambda: None)
    e.two_int.connect(lambda x: None)
    assert len(e.two_int._slots) == 2
    e.two_int.connect(lambda x, y: None)
    e.two_int.connect(lambda x, y, z=None: None)
    assert len(e.two_int._slots) == 4
    e.two_int.connect(lambda x, y, *z: None)
    e.two_int.connect(lambda *z: None)
    assert len(e.two_int._slots) == 6
    e.two_int.connect(lambda *z, **k: None)
    assert len(e.two_int._slots) == 7

    with pytest.raises(ValueError):
        e.two_int.connect(lambda x, y, z: None)


def test_mock_connect():
    e = Emitter()
    e.one_int.connect(MagicMock())


# fmt: off
class TypeA: ...
class TypeB(TypeA): ...
class TypeC(TypeB): ...
class Rcv:
    def methodA(self, obj: TypeA): ...
    def methodA_ref(self, obj: 'TypeA'): ...
    def methodB(self, obj: TypeB): ...
    def methodB_ref(self, obj: 'TypeB'): ...
    def methodOptB(self, obj: TypeB | None): ...
    def methodOptB_ref(self, obj: 'TypeB | None'): ...
    def methodC(self, obj: TypeC): ...
    def methodC_ref(self, obj: 'TypeC'): ...
class Emt:
    signal = Signal(TypeB)
# fmt: on


def test_forward_refs_type_checking():
    e = Emt()
    r = Rcv()
    e.signal.connect(r.methodB, check_types=True)
    e.signal.connect(r.methodB_ref, check_types=True)
    e.signal.connect(r.methodOptB, check_types=True)
    e.signal.connect(r.methodOptB_ref, check_types=True)
    e.signal.connect(r.methodC, check_types=True)
    e.signal.connect(r.methodC_ref, check_types=True)

    # signal is emitting a TypeB, but method is expecting a typeA
    assert not issubclass(TypeA, TypeB)
    # typeA is not a TypeB, so we get an error

    with pytest.raises(ValueError):
        e.signal.connect(r.methodA, check_types=True)
    with pytest.raises(ValueError):
        e.signal.connect(r.methodA_ref, check_types=True)


def test_checking_off():
    e = Emitter()

    # the no_check signal was instantiated with check_[nargs/types] = False
    @e.no_check.connect
    def bad_in_many_ways(x: int, y, z): ...


def test_keyword_only_not_allowed():
    e = Emitter()

    def f(a: int, *, b: int): ...

    with pytest.raises(ValueError) as er:
        e.two_int.connect(f)
    assert "Unsupported KEYWORD_ONLY parameters in signature" in str(er)


def test_unique_connections():
    e = Emitter()
    assert len(e.one_int._slots) == 0

    e.one_int.connect(f_no_arg, unique=True)
    assert len(e.one_int._slots) == 1

    e.one_int.connect(f_no_arg, unique=True)
    assert len(e.one_int._slots) == 1

    with pytest.raises(ValueError):
        e.one_int.connect(f_no_arg, unique="raise")
    assert len(e.one_int._slots) == 1

    e.one_int.connect(f_no_arg)
    assert len(e.one_int._slots) == 2


def test_sig_unavailable():
    """In some cases, signature.inspect() fails on a callable, (many builtins).

    We should still connect, but with a warning.
    """
    e = Emitter()
    e.one_int.connect(vars, check_nargs=False)  # no warning

    with pytest.warns(UserWarning):
        e.one_int.connect(vars)

    # we've special cased print... due to frequency of use.
    e.one_int.connect(print)  # no warning


def test_pause():
    """Test that we can pause, and resume emission of (possibly reduced) args."""
    emitter = Emitter()
    mock = MagicMock()

    emitter.one_int.connect(mock)
    emitter.one_int.emit(1)
    mock.assert_called_once_with(1)

    mock.reset_mock()
    emitter.one_int.pause()
    emitter.one_int.emit(1)
    emitter.one_int.emit(2)
    emitter.one_int.emit(3)
    mock.assert_not_called()
    emitter.one_int.resume()
    mock.assert_has_calls([call(1), call(2), call(3)])

    mock.reset_mock()
    with emitter.one_int.paused(lambda a, b: (a[0].union(set(b)),), (set(),)):
        emitter.one_int.emit(1)
        emitter.one_int.emit(2)
        emitter.one_int.emit(3)
    mock.assert_called_once_with({1, 2, 3})

    mock.reset_mock()
    emitter.one_int.pause()
    emitter.one_int.resume()
    mock.assert_not_called()


def test_resume_with_initial():
    emitter = Emitter()
    mock = MagicMock()
    emitter.one_int.connect(mock)

    with emitter.one_int.paused(lambda a, b: (a[0] + b[0],)):
        emitter.one_int.emit(1)
        emitter.one_int.emit(2)
        emitter.one_int.emit(3)
    mock.assert_called_once_with(6)

    mock.reset_mock()
    with emitter.one_int.paused(lambda a, b: (a[0] + b[0],), (20,)):
        emitter.one_int.emit(1)
        emitter.one_int.emit(2)
        emitter.one_int.emit(3)
    mock.assert_called_once_with(26)


def test_nested_pause():
    emitter = Emitter()
    mock = MagicMock()
    emitter.one_int.connect(mock)
    with emitter.one_int.paused():
        emitter.one_int.emit(1)
        emitter.one_int.emit(2)
        with emitter.one_int.paused():
            emitter.one_int.emit(3)
            emitter.one_int.emit(4)
        emitter.one_int.emit(5)
    mock.assert_has_calls([call(i) for i in (1, 2, 3, 4, 5)])


def test_signals_on_unhashables():
    class Emitter(dict):
        signal = Signal(int)

    e = Emitter()
    e.signal.connect(lambda x: print(x))
    e.signal.emit(1)


def test_property_connect():
    class A:
        def __init__(self):
            self.li = []

        @property
        def x(self):
            return self.li

        @x.setter
        def x(self, value):
            self.li.append(value)

    a = A()
    emitter = Emitter()

    emitter.one_int.connect_setattr(a, "x", maxargs=1)
    assert len(emitter.one_int) == 1
    with pytest.warns(FutureWarning, match="The default value of maxargs"):
        emitter.two_int.connect_setattr(a, "x")
    assert len(emitter.two_int) == 1
    emitter.one_int.emit(1)
    assert a.li == [1]
    emitter.two_int.emit(1, 1)
    assert a.li == [1, (1, 1)]
    emitter.two_int.disconnect_setattr(a, "x")
    assert len(emitter.two_int) == 0
    with pytest.raises(ValueError):
        emitter.two_int.disconnect_setattr(a, "x", missing_ok=False)
    emitter.two_int.disconnect_setattr(a, "x")
    s = emitter.two_int.connect_setattr(a, "x", maxargs=1)
    emitter.two_int.emit(2, 3)
    assert a.li == [1, (1, 1), 2]
    emitter.two_int.disconnect(s, missing_ok=False)

    with pytest.raises(AttributeError):
        emitter.one_int.connect_setattr(a, "y", maxargs=None)


def test_connect_setitem():
    class T:
        sig = Signal(int)

    class SupportsItem:
        def __init__(self) -> None:
            self._dict = {}

        def __setitem__(self, key, value):
            self._dict[key] = value

    t = T()
    my_obj = SupportsItem()
    with pytest.warns(FutureWarning, match="The default value of maxargs"):
        t.sig.connect_setitem(my_obj, "x")
    t.sig.emit(5)
    assert my_obj._dict == {"x": 5}
    t.sig.disconnect_setitem(my_obj, "x")
    t.sig.emit(7)
    assert my_obj._dict == {"x": 5}

    obj = object()
    with pytest.raises(TypeError, match="does not support __setitem__"):
        t.sig.connect_setitem(obj, "x", maxargs=1)

    with pytest.raises(TypeError):
        t.sig.disconnect_setitem(obj, "x", missing_ok=False)


def test_repr_not_used():
    """Test that we don't use repr() or __call__ to check signature."""
    mock = MagicMock()

    class T:
        def __repr__(self):
            mock()
            return "<REPR>"

        def __call__(self):
            mock()

    t = T()
    sig = SignalInstance()
    sig.connect(t)
    mock.assert_not_called()


# b.signal2.emit will warn that compiled SignalInstances cannot be weakly referenced
@pytest.mark.filterwarnings("ignore:failed to create weakref:UserWarning")
def test_signal_emit_as_slot():
    class A:
        signal1 = Signal(int)

    class B:
        signal2 = Signal(int)

    mock = Mock()
    a = A()
    b = B()
    a.signal1.connect(b.signal2.emit)
    b.signal2.connect(mock)
    a.signal1.emit(1)
    mock.assert_called_once_with(1)

    mock.reset_mock()
    a.signal1.disconnect(b.signal2.emit)
    a.signal1.connect(b.signal2)  # you can also just connect the signal instance
    a.signal1.emit(2)
    mock.assert_called_once_with(2)


def test_emit_loop_exceptions():
    emitter = Emitter()
    mock1 = Mock(side_effect=ValueError("Bad callback!"))
    mock2 = Mock()
    emitter.one_int.connect(mock1)
    emitter.one_int.connect(mock2)

    with pytest.raises(EmitLoopError):
        emitter.one_int.emit(1)

    mock1.assert_called_once_with(1)
    mock1.reset_mock()
    mock2.assert_not_called()

    with suppress(EmitLoopError):
        emitter.one_int.emit(2)
    mock1.assert_called_once_with(2)
    mock1.assert_called_once_with(2)


@pytest.mark.parametrize(
    "slot",
    [
        "f_no_arg",
        "f_int_decorated_stupid",
        "f_int_decorated_good",
        "f_any_assigned",
        "partial",
        "partial_kwargs",
        "partial",
    ],
)
def test_weakref_disconnect(slot):
    """Test that a connected method doesn't hold strong ref."""

    emitter = Emitter()
    obj = MyObj()

    assert len(emitter.one_int) == 0
    if slot == "partial":
        cb = partial(obj.f_int_int, 1)
    elif slot == "partial_kwargs":
        cb = partial(obj.f_int_int, b=1)
    else:
        cb = getattr(obj, slot)
    emitter.one_int.connect(cb)
    assert len(emitter.one_int) == 1
    emitter.one_int.emit(1)
    assert len(emitter.one_int) == 1
    emitter.one_int.disconnect(cb)
    assert len(emitter.one_int) == 0


def test_queued_connections():
    from threading import Thread, current_thread

    from psygnal import emit_queued

    this_thread = current_thread()

    emitter = Emitter()

    # function to run in another thread
    def _run():
        emit_queued()
        emitter.one_int.emit(2)

    other_thread = Thread(target=_run)

    this_thread_mock = Mock()
    other_thread_mock = Mock()
    any_thread_mock = Mock()

    # mock1 wants to be called in this thread
    @emitter.one_int.connect(thread=this_thread)
    def cb1(arg):
        this_thread_mock(arg, current_thread())

    # mock2 wants to be called in other_thread
    @emitter.one_int.connect(thread=other_thread)
    def cb2(arg):
        other_thread_mock(arg, current_thread())

    # mock3 wants to be called in whatever thread the emitter is in
    @emitter.one_int.connect
    def cb3(arg):
        any_thread_mock(arg, current_thread())

    # emit in this thread
    emitter.one_int.emit(1)
    this_thread_mock.assert_called_once_with(1, this_thread)
    # other_thread_mock not called because it's waiting for other_thread
    other_thread_mock.assert_not_called()
    # any_thread_mock called because it's waiting for any thread
    any_thread_mock.assert_called_once_with(1, this_thread)

    # Now we run `_run` in other_thread
    this_thread_mock.reset_mock()
    any_thread_mock.reset_mock()
    other_thread.start()
    other_thread.join()

    # now mock2 should be called TWICE.  Once for the .emit(1) queued from this thread,
    # and once for the .emit(2) in other_thread
    other_thread_mock.assert_has_calls([call(1, other_thread), call(2, other_thread)])
    # stuff queued for any_thread_mock should have also been called
    any_thread_mock.assert_called_once_with(2, other_thread)

    # stuff queued for this thread should NOT have been called
    this_thread_mock.assert_not_called()
    # ... until we call emit_queued() from this thread
    emit_queued()
    this_thread_mock.assert_called_once_with(2, this_thread)


def test_deepcopy():
    from copy import deepcopy

    mock = Mock()

    class T:
        sig = Signal()

    t = T()

    @t.sig.connect
    def f():
        mock()

    t.sig.emit()
    mock.assert_called_once()
    mock.reset_mock()

    x = deepcopy(t)
    assert x is not t
    x.sig.emit()
    mock.assert_called_once()

    mock2 = Mock()

    class Foo:
        def method(self):
            mock2()

    foo = Foo()
    t.sig.connect(foo.method)
    t.sig.emit()
    mock2.assert_called_once()
    mock2.reset_mock()

    with pytest.warns(UserWarning, match="does not copy connected weakly referenced"):
        x2 = deepcopy(t)

    x2.sig.emit()
    mock2.assert_not_called()


class T:
    sig = Signal()


mock = Mock()


def f():
    return mock()


def test_pickle():
    import pickle

    t = T()

    t.sig.connect(f)

    _dump = pickle.dumps(t)
    x = pickle.loads(_dump)
    x.sig.emit()
    mock.assert_called_once()


@pytest.mark.parametrize("strategy", [ReemissionMode.QUEUED, ReemissionMode.IMMEDIATE])
def test_recursion_error(strategy: ReemissionMode) -> None:
    s = SignalInstance(reemission=strategy)

    @s.connect
    def callback() -> None:
        s.emit()

    with pytest.raises(RecursionError):
        s.emit()


@pytest.mark.parametrize(
    "strategy", [ReemissionMode.QUEUED, ReemissionMode.IMMEDIATE, ReemissionMode.LATEST]
)
def test_callback_order(strategy: ReemissionMode) -> None:
    sig = SignalInstance((int,), reemission=strategy)

    a = []

    def cb1(value: int) -> None:
        a.append(value)
        if value == 1:
            sig.emit(2)

    def cb2(value: int) -> None:
        a.append(value * 10)
        if value == 2:
            sig.emit(3)

    def cb3(value: int) -> None:
        a.append(value * 100)

    sig.connect(cb1)
    sig.connect(cb2)
    sig.connect(cb3)
    sig.emit(1)

    if strategy == ReemissionMode.IMMEDIATE:
        # nested emission events immediately trigger the next nested level
        # before returning to process the remainder of the current loop
        assert a == [1, 2, 20, 3, 30, 300, 200, 10, 100]
    elif strategy == ReemissionMode.LATEST:
        # nested emission events immediately trigger the next nested level
        # and never return to process the remainder of the current loop
        assert a == [1, 2, 20, 3, 30, 300]
    elif strategy == ReemissionMode.QUEUED:
        # all callbacks are called once before the next one is called
        assert a == [1, 10, 100, 2, 20, 200, 3, 30, 300]


@pytest.mark.parametrize("strategy", [ReemissionMode.QUEUED, ReemissionMode.IMMEDIATE])
def test_signal_order_suspend(strategy: ReemissionMode) -> None:
    """Test that signals are emitted in the order they were connected."""
    sig = SignalInstance((int,), reemission=strategy)
    mock1 = Mock()
    mock2 = Mock()

    def callback(x):
        if x < 10:
            sig.emit(10)

    def callback2(x):
        if x == 10:
            sig.emit(11)

    def callback3(x):
        if x == 10:
            with sig.paused(reducer=lambda a, b: (a[0] + b[0],)):
                for i in range(12, 15):
                    sig.emit(i)

    sig.connect(mock1)
    sig.connect(callback)
    sig.connect(callback2)
    sig.connect(callback3)
    sig.connect(mock2)
    sig.emit(1)

    mock1.assert_has_calls([call(1), call(10), call(11), call(39)])
    if strategy == ReemissionMode.IMMEDIATE:
        mock2.assert_has_calls([call(11), call(39), call(10), call(1)])
    elif strategy == ReemissionMode.QUEUED:
        mock2.assert_has_calls([call(1), call(10), call(11), call(39)])


def test_call_priority() -> None:
    """Test that signals are emitted in the order they were connected."""
    emitter = Emitter()

    calls = []

    emitter.no_arg.connect(lambda: calls.append(1))
    emitter.no_arg.connect(lambda: calls.append(2), priority=5)
    emitter.no_arg.connect(lambda: calls.append(3), priority=-5)
    emitter.no_arg.connect(lambda: calls.append(4), priority=10)

    emitter.no_arg.emit()
    assert calls == [4, 2, 1, 3]


def test_slotted_classes() -> None:
    class T:
        __slots__ = ("not_sig",)
        sig = Signal()

    t = T()
    mock = Mock()

    @t.sig.connect
    def f():
        mock()

    t.sig.emit()
    mock.assert_called_once()

    assert t.sig is t.sig


def test_emit_should_not_prevent_gc():
    from weakref import WeakSet

    from psygnal import Signal

    class Obj:
        pass

    class SomethingWithSignal:
        changed = Signal(object)

    object_instances: WeakSet[Obj] = WeakSet()
    something = SomethingWithSignal()
    obj = Obj()
    object_instances.add(obj)
    something.changed.emit(obj)
    del obj
    assert len(object_instances) == 0


@pytest.mark.parametrize("strategy", ReemissionMode._members())
def test_emit_loop_error_message_construction(strategy: ReemissionVal) -> None:
    sig = SignalInstance((int,), reemission=strategy)
    sig.connect(lambda v: v == 1 and sig.emit(2))  # type: ignore
    sig.connect(lambda v: v == 2 and sig.emit(0))  # type: ignore
    sig.connect(lambda v: 1 / v)
    with pytest.raises(EmitLoopError, match="While emitting signal") as e:
        sig.emit(1)

    if strategy == "queued":
        # check that we show a useful message for confusign queued signals
        assert "NOTE" in str(e.value)


def test_description():
    description = "A signal"

    class T:
        sig = Signal(description=description)

    assert T.sig.description == description
    assert T().sig.description == description
